"""
LS machine learning integration. Useful resources:

- Possible routes: https://github.com/HumanSignal/label-studio-ml-backend/blob/master/label_studio_ml/api.py

The instruction in your project can be as an example:

<style>
.ls-modal__content {
    width: auto;
    height: "800px";
}
</style>
This needs to be reachable from the local user machine:
<iframe src="http://localhost:8000/lgbm/settingsgui" width="100%" height="700px"></iframe>
"""

import json
import base64
import logging
import jinja2
from typing import Any, TypeVar, Generic, Literal
import numpy as np

from mmm.BaseModel import BaseModel
from .projects import LSProject
from .LabelstudioCredentials import LabelstudioCredentials

try:
    # API extras
    from label_studio_sdk import Client, Project
    from label_studio_converter.brush import (
        decode_rle,
        encode_rle,
        decode_from_annotation,
    )
    from fastapi import APIRouter, BackgroundTasks, Request
    from fastapi.responses import HTMLResponse
except ImportError:
    Client, Project = Any, Any
    APIRouter, BackgroundTasks, Request, HTMLResponse = Any, Any, Any, Any

# bound to be a pydantic model
T = TypeVar("T", bound=BaseModel)


class LabelStudioTask(BaseModel):
    data: dict
    predictions: list[dict] | None = None
    annotations: list[dict] | None = None
    meta: dict | None = None


class SerializedArray(BaseModel):
    """
    >>> import numpy as np
    >>> from mmm.labelstudio_ext.LSModel import SerializedArray
    >>> for dtype in [np.int8, np.int32, np.int64, np.float32, np.float64]:
    ...     testarr = np.random.rand(50, 52).astype(dtype)
    ...     seria_arr = SerializedArray.from_numpy(testarr)
    ...     assert np.array_equal(testarr, seria_arr.to_numpy())
    """

    data: str
    dtype: str
    shape: tuple[int, ...]
    meta: dict | None = None

    @classmethod
    def from_numpy(cls, arr: np.ndarray, meta: dict | None = None):
        return cls(
            data=base64.b64encode(arr).decode("utf-8"),
            dtype=str(arr.dtype),
            shape=arr.shape,
            meta=meta,
        )

    def to_numpy(self):
        return np.frombuffer(base64.b64decode(self.data), dtype=self.dtype).reshape(self.shape)


class LSModel(Generic[T]):
    """
    For predict to be called,
    select "Retrieve predictions when loading a task automatically" in the Machine learning settings.

    Train gets called by Labelstudio when the user clicks "Train" in the UI.
    Alternatively, enable the setting: "Start model training after any annotation are submitted or updated".
    """

    class Config(BaseModel):
        category: str = "unknownmodel"

    def __init__(self, cfg: T, ls_client: Client, prefix: str) -> None:
        self.cfg, self.ls_client, self.prefix = cfg, ls_client, prefix
        self.mtl_templates = jinja2.Environment(loader=jinja2.PackageLoader("mmm", "resources"))
        self.model_version = None

    async def get_user_helper(self) -> str:
        return f"Please configure model {self.cfg.category} hosted at {self.prefix}"

    async def get_user_html(self) -> str:
        return f"<h1>Model {self.cfg.category} hosted at {self.prefix}</h1>"

    async def settings_gui(self, request: Request) -> str:
        # Render the jsonform.jinja2 template
        return HTMLResponse(
            content=self.mtl_templates.get_template("jsonform.jinja2").render(
                {
                    "helper": await self.get_user_helper(),
                    "post_url": f"{self.prefix}/settings",
                    "data": (await self.get_current_settings()).model_dump_json(),
                    "schema": json.dumps(await self.get_settings_schema()),
                    "user_html": await self.get_user_html(),
                }
            ),
            status_code=200,
        )

    async def get_settings_schema(self) -> dict:
        return self.cfg.model_json_schema()

    async def get_current_settings(self) -> T:
        return self.cfg

    async def set_settings(self, body: dict) -> T:
        self.cfg = type(self.cfg)(**body)
        return self.cfg

    async def predict(self, project: LSProject, tasks: list[int], context: dict | None = None):
        """
        If this is called with a single task, we assume that the respective user just clicked on this task.

        https://labelstud.io/api#tag/Predictions/operation/api_predictions_create
        """
        logging.info(f"Predict not implemented, called for project {project} and {len(tasks)} tasks")
        return {}

    async def train(
        self,
        project: LSProject,
        tasks: list[dict],
        mode: Literal["retrain", "continue"] = "retrain",
    ):
        logging.info(f"Train not implemented, called for project {project} and {len(tasks)} tasks with mode {mode}")
        return {}

    async def get_version(self):
        if self.model_version is not None:
            return {"model_version": self.model_version}
        return {"model_version": "unknown"}

    async def get_versions(self):
        d = await self.get_version()
        return {"versions": [d["model_version"]]}

    async def _train_handler(self, request: Request):
        requestbody = await request.json()
        project_id: str = requestbody["project"].split(".")[0]
        ls_project: Project = self.ls_client.get_project(project_id)
        project: LSProject = LSProject(
            LSProject.Config(name=ls_project.title),
            self.ls_client,
            ls_project=ls_project,
        )
        return await self.train(
            project,
            requestbody["annotations"] if "annotations" in requestbody else [],
            mode="retrain",
        )

    async def _predict_handler(self, request: Request):
        req = await request.json()
        project_id = req["project"].split(".")[0]
        context = req["params"]["context"]
        if context is not None:
            logging.info(context)
        ls_project: Project = self.ls_client.get_project(project_id)
        project: LSProject = LSProject(
            LSProject.Config(name=ls_project.title),
            self.ls_client,
            ls_project=ls_project,
        )
        logging.info(f"Called predict with {len(req['tasks'])} tasks: {[t['id'] for t in req['tasks']]}")
        predictions = await self.predict(project, req["tasks"], context=context)
        return predictions

    async def _setup_handler(self, request: Request):
        req_body = await request.json()
        project_id: str = req_body["project"].split(".")[0]
        ls_project = self.ls_client.get_project(project_id)
        instruction = ls_project.params["expert_instruction"]
        self.model_version = req_body["model_version"]

        try:
            f_key, f_instr = await self.get_instruction(ls_project, instruction)

            if not f_key(instruction):
                logging.info(f"Adding instruction for {self.prefix}")
                ls_project.set_params(**{"expert_instruction": f_instr(ls_project, instruction)})
        except NotImplementedError:
            logging.info(f"Instruction for {self.prefix} not implemented")

        return await self.get_version()

    async def get_instruction(self, project: Project, previous_instruction: str) -> tuple:
        """
        First function is the key that is used to see if the instruction is already there.
        Second function builds the instruction itself.
        """
        raise NotImplementedError()

    async def webhook(self, request: Request, background_tasks: BackgroundTasks):
        body = await request.json()
        if body["action"] in ["ANNOTATION_UPDATED", "ANNOTATION_CREATED"]:
            logging.info(f"Annotation updated: {body['task']['id']}")
            project_id: str = body["project"]["id"]
            ls_project: Project = self.ls_client.get_project(project_id)
            project: LSProject = LSProject(
                LSProject.Config(name=ls_project.title),
                self.ls_client,
                ls_project=ls_project,
            )
            ls_task = body["task"]
            ls_task["annotations"] = [body["annotation"]]
            background_tasks.add_task(self.train, project, [ls_task], mode="continue")
        return {}

    async def health_handler(self):
        return {"status": "UP", "model_class": self.cfg.category}

    def get_tag(self):
        return f"Model {self.cfg.category}"

    async def invocation(self, ls_task: LabelStudioTask) -> SerializedArray:
        """
        Enables to perform inference without requiring a labelstudio project.
        """
        raise NotImplementedError

    async def labeling_priority(self, ls_task: LabelStudioTask) -> float:
        """
        Returns a high number for tasks that should be labeled.
        """
        raise NotImplementedError

    def build_router(self) -> APIRouter:
        router = APIRouter()
        tag = self.get_tag()
        # Required for labelstudio
        labelstudio_tag = f"Labelstudio {self.cfg.category}"
        router.post("/predict", tags=[labelstudio_tag])(self._predict_handler)
        router.post("/train", tags=[labelstudio_tag])(self._train_handler)
        router.get("/health", tags=[labelstudio_tag])(self.health_handler)
        router.post("/webhook", tags=[labelstudio_tag])(self.webhook)
        router.post("/setup", tags=[labelstudio_tag])(self._setup_handler)
        router.get("/versions", tags=[labelstudio_tag])(self.get_versions)

        # User interaction
        router.get("/settingsgui", tags=[labelstudio_tag])(self.settings_gui)
        router.get("/settingsschema", tags=[labelstudio_tag])(self.get_settings_schema)
        router.get("/settings", tags=[labelstudio_tag])(self.get_current_settings)
        router.post("/settings", tags=[labelstudio_tag])(self.set_settings)

        # Invocation without labelstudio
        router.post("/compute_score/", tags=[tag])(self.labeling_priority)
        router.post("/invocation", tags=[tag])(self.invocation)

        return router



"""
Wrappers for models imported from torchvision model zoo
"""

from typing import List, Dict, Literal, Tuple
from pydantic import Field

import torch.nn as nn

import torch
import torch.nn as nn
import torchvision.models as torch_models
from torchvision.models._utils import IntermediateLayerGetter
import logging

from ..TorchModule import TorchModule
from mmm.torch_ext import infer_stride_channels_from_features


class FeatureBackbone(nn.Module):
    def __init__(self, backbone: nn.Module, return_layers: Dict[str, str]) -> None:
        super().__init__()

        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        raw_dict = self.body(x)

        return list(raw_dict.values())


layername_KB = {
    "densenet121": {
        "constr": lambda w: torch_models.densenet121(weights=w).features,
        "layernames": ["denseblock1", "denseblock2", "denseblock3", "denseblock4"],
        "weights": torch_models.DenseNet121_Weights.IMAGENET1K_V1,
    },
    "resnet18": {
        "constr": lambda w: torch_models.resnet18(weights=w),
        "layernames": ["layer1", "layer2", "layer3", "layer4"],
        "weights": torch_models.ResNet18_Weights.DEFAULT,
    },
    "resnet50": {
        "constr": lambda w: torch_models.resnet50(weights=w),
        "layernames": ["layer1", "layer2", "layer3", "layer4"],
        "weights": torch_models.ResNet50_Weights.DEFAULT,
    },
    "efficientnet_v2_s": {
        "constr": lambda w: torch_models.efficientnet_v2_s(weights=w).features,
        "layernames": ["2", "3", "5", "7"],  # hand picked for strides 4, 8, 16, 32
        "weights": torch_models.EfficientNet_V2_S_Weights.DEFAULT,
    },
}


class TorchVisionCNN(nn.Module):
    class Config(TorchModule):
        architecture: Literal["residualcnn"] = "residualcnn"
        variant: Literal[tuple(layername_KB.keys())] = "densenet121"  # type: ignore
        pretrained: bool = Field(default=True, description="Will load ImageNet1k weights from TorchVision")

        def build_instance(self, *args, **kwargs) -> nn.Module:
            return TorchVisionCNN(self)

    def __init__(self, args: Config) -> None:
        super().__init__()
        self.args = args

        kb = layername_KB[self.args.variant]

        weights = kb["weights"] if self.args.pretrained else None

        self.backbone = FeatureBackbone(
            # torch_models.densenet121(weights=weights).features,
            kb["constr"](weights),
            # return_layers={v: v for v in ["denseblock1", "denseblock2", "denseblock3", "denseblock4"]},
            return_layers={v: v for v in kb["layernames"]},
        )

        with torch.no_grad():
            self.channels, self.strides = infer_stride_channels_from_features(self(torch.rand(1, 3, 224, 224)))

    def get_feature_pyramid_channels(self) -> List[int]:
        # return [3] + self.backbone.in_channels_list
        return self.channels

    def get_strides(self):
        # Input and spatial sizes of the feature maps
        return self.strides

    def forward(self, x):
        pyr = self.backbone(x)
        return [x] + pyr



import torch
import json
import numpy as np
from typing import Any
from pathlib import Path
from torch.utils.data import Dataset
import torchvision.transforms.functional as F
import imageio.v3 as imageio
from torchvision.datasets.folder import default_loader
from mmm.BaseModel import BaseModel
from mmm.utils import disk_cacher

# Enable extra fields such that not everything from labelstudio needs to be modeled
from pydantic import BaseModel as RawBaseModel, ConfigDict

from .projects import LSProject
from mmm.labelstudio_ext.utils import download_image, convert_task_to_seglabel


class LSCase(RawBaseModel):  # Not use mmm.BaseModel to allow for extra fields
    model_config = ConfigDict(extra="allow")
    id: int
    task: dict
    project: int
    result: list[dict]


class LSDataset(Dataset):
    """
    Wraps a target storage of labelstudio
    """

    class Config(BaseModel):
        abs_folder_path: str
        # only_for_project_id: None | str = None
        project: LSProject.Config = LSProject.Config(name="syndifact")

    def __init__(self, args: Config):
        self.cfg = args
        self.folder_path = Path(self.cfg.abs_folder_path)
        self.case_paths: list[Path] = self.update_from_disk()

    def update_from_disk(self) -> list[Path]:
        self.case_paths = list(self.folder_path.iterdir())

        return self.case_paths

    def __len__(self):
        self.update_from_disk()
        return len(self.case_paths)

    def __getitem__(self, index) -> dict:
        case_path = self.case_paths[index]
        case_dict = json.loads((case_path).read_text())
        return LSCase(**case_dict)

    @staticmethod
    def loadmcsemseg(x: LSCase, project: LSProject):
        """
        can be used as a src_transform like `partial(loadmcsemseg, project=ls_dataset.project)`
        """

        @disk_cacher(cache_path="shared")
        def download_image_cached(url):
            """
            Labelstudio often only stores the url to the image, so we need to download it.
            During training that can be a bottleneck, so we cache it by default.
            """
            return download_image(url)

        img = download_image_cached(x.task["data"]["image"])
        seglabel = convert_task_to_seglabel(
            {
                "id": x.task["id"],
                "annotations": [{"result": x.result}],
                "data": {"image": x.task["data"]["image"]},
                "meta": {"labelstudioid": x.task["id"]},
            },
            project.ls.parsed_label_config["histoarteseg"]["labels"],
            prefill_value=-1,
        )
        return {
            "image": F.to_tensor(img),
            "label": torch.from_numpy(seglabel).long(),
            "meta": {
                "labelstudioid": x.task["id"],
                "lslink": f"/projects/{project.ls.id}/data?task={x.task['id']}",
            },
        }



import random
from typing import Literal, Iterable
from pydantic import Field
import itertools

from shapely.geometry import Polygon
from shapely import GEOSException

from mmm.BaseModel import BaseModel

from .GeoAnno import GeoAnno
from .NoUsefulWindowException import NoUsefulWindowException


class GeojsonRegionWindows(BaseModel):
    """
    Yields subwindows inside annotated regions given as coarse polygons.
    """

    min_area: float = Field(
        default=0.95,
        description="Area of a window that is required to be inside the annotated region. 0.95 for 95%",
    )
    patch_size: tuple[int, int] = 224, 224
    coordinate_augmentation: Literal["random", "none"] = "random"
    windowsize_augmentation: Literal["relative", "none"] = "relative"
    augmentation_strength: float = Field(
        default=0.2,
        description="Controls the windowsize augmentation. Should be smaller than 1.",
    )
    stepsize: int = Field(
        default=2,
        description="Controls the stepsize. For one moves one window at a time, for 2 it moves half a window every step",
    )

    def augment_window(self, x_raw, y_raw, l0_window_width, l0_window_height):
        """
        If corresponding config options are set, will jiggle coordinates and window size.
        """
        if self.coordinate_augmentation == "random":
            rand_x = int(l0_window_width // (self.stepsize * 2))
            rand_y = int(l0_window_height // (self.stepsize * 2))
            x = x_raw + random.randint(-1 * rand_x, rand_x)
            y = y_raw + random.randint(-1 * rand_y, rand_y)
        elif self.coordinate_augmentation == "none":
            x, y = x_raw, y_raw
        else:
            raise NotImplementedError(f"Unknown stepsize_method {self.coordinate_augmentation}")

        if self.windowsize_augmentation == "relative":
            rand_width = int(l0_window_width * self.augmentation_strength)
            rand_height = int(l0_window_height * self.augmentation_strength)
            width = l0_window_width + random.randint(-1 * rand_width, rand_width)
            height = l0_window_height + random.randint(-1 * rand_height, rand_height)
        elif self.windowsize_augmentation == "none":
            width, height = l0_window_width, l0_window_height
        else:
            raise NotImplementedError(f"Unknown windowsize augmentation method {self.windowsize_augmentation}")

        return Polygon([(x, y), (x + width, y), (x + width, y + height), (x, y + height)])

    def test_window_validity(self, checkwindow: Polygon, anno: GeoAnno) -> bool:
        try:
            if checkwindow.intersects(anno.shape):
                if (checkwindow.intersection(anno.shape).area / checkwindow.area) >= self.min_area:
                    return True
        except GEOSException:
            return False
        return False

    def iter_valid_windows(
        self, anno: GeoAnno, level_downsamples: dict[int, float]
    ) -> Iterable[tuple[int, tuple[int, int], tuple[int, int]]]:
        """
        Yields all windows with sufficient area inside `anno`.

        Uses a generator. As a result you can e.g. use `itertools.islice` to limit the number of windows per annotation.

        The patch_size is given. In consequence, only (level, location) tuples are yielded.
        """
        window_sizes = {
            l: (
                self.patch_size[0] * downsample_fac,
                self.patch_size[1] * downsample_fac,
            )
            for l, downsample_fac in level_downsamples.items()
        }

        # It does not make sense if a window is larger than the region annotation itself
        useful_window_sizes = {k: v for k, v in window_sizes.items() if v[0] * v[1] <= anno.shape.area}
        if not useful_window_sizes:
            raise NoUsefulWindowException(
                f"Annotation did not have a useful window in levels {level_downsamples}: {anno}"
            )
        proposal_coords = {}
        for l, (l0_window_width, l0_window_height) in useful_window_sizes.items():
            xcoords = list(
                range(
                    int(anno.shape.bounds[0] - l0_window_width),
                    int(anno.shape.bounds[2]),
                    int(l0_window_width // self.stepsize),
                )
            )
            ycoords = list(
                range(
                    int(anno.shape.bounds[1] - l0_window_height),
                    int(anno.shape.bounds[3]),
                    int(l0_window_height // self.stepsize),
                )
            )

            # Itertools.product does not yield random combinations, so a list conversion is necessary
            cs = list(itertools.product(xcoords, ycoords))
            random.shuffle(cs)
            proposal_coords[l] = (coord for coord in cs)

        valid_levels: list[int] = list(useful_window_sizes.keys())
        # current_level_idx = 0  # in first iteration it will be 0
        while valid_levels:
            selected_level = random.choice(valid_levels)
            try:
                # selected_level = valid_levels[current_level_idx]
                x_raw, y_raw = next(proposal_coords[selected_level])
                l0_window_width, l0_window_height = useful_window_sizes[selected_level]
                l0_window: Polygon = self.augment_window(x_raw, y_raw, l0_window_width, l0_window_height)
                if self.test_window_validity(l0_window, anno):
                    x1, y1, x2, y2 = l0_window.bounds
                    level_patchsize = (
                        int((x2 - x1) / level_downsamples[selected_level]),
                        int((y2 - y1) / level_downsamples[selected_level]),
                    )
                    yield selected_level, (int(x1), int(y1)), level_patchsize
                    # current_level_idx = (current_level_idx + 1) % len(valid_levels)
            except StopIteration:
                # No more proposal coordinates in current_level
                valid_levels.remove(selected_level)
                # current_level_idx = 0



from typing import List
import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn

from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.PyramidEncoder import PyramidEncoder


class SCSEModule(nn.Module):
    """
    Adapted from SMP
    """

    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.cSE = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, in_channels // reduction, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels // reduction, in_channels, 1),
            nn.Sigmoid(),
        )
        self.sSE = nn.Sequential(nn.Conv2d(in_channels, 1, 1), nn.Sigmoid())

    def forward(self, x):
        return x * self.cSE(x) + x * self.sSE(x)


class Attention(nn.Module):
    """
    Adapted from SMP
    """

    def __init__(self, name, **params):
        super().__init__()

        if name is None:
            self.attention = nn.Identity(**params)
        elif name == "scse":
            self.attention = SCSEModule(**params)
        else:
            raise ValueError("Attention {} is not implemented".format(name))

    def forward(self, x):
        return self.attention(x)


class Conv2dReLU(nn.Sequential):
    """
    Adapted from SMP
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        padding=1,
        stride=1,
    ):
        conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=False,
        )
        relu = nn.ReLU(inplace=True)

        # Batchnorm2d needs to be used here, the Sharedblock class will replace by the correct norm
        bn = nn.BatchNorm2d(out_channels)

        super(Conv2dReLU, self).__init__(conv, bn, relu)


class DecoderBlock(nn.Module):
    """
    Adapted from SMP.

    If attention type is None, the attention module is just an identity.
    """

    def __init__(
        self,
        in_channels,
        skip_channels,
        out_channels,
        attention_type=None,
        third_conv=True,
    ):
        super().__init__()
        self.conv1 = Conv2dReLU(
            in_channels + skip_channels,
            out_channels,
            kernel_size=3,
            padding=1,
        )
        self.attention1 = Attention(attention_type, in_channels=in_channels + skip_channels)
        self.conv2 = Conv2dReLU(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
        )
        self.attention2 = Attention(attention_type, in_channels=out_channels)
        if third_conv:
            self.conv3 = Conv2dReLU(
                out_channels,
                out_channels,
                kernel_size=3,
                padding=1,
            )
        else:
            self.conv3 = nn.Identity()

    def forward(self, x, skip=None):
        """
        skip is the large feature map that x will be concatenated to.
        """
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
            x = self.attention1(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.attention2(x)
        return x


class LatentBlock(nn.Module):
    def __init__(self, hidden_dim, out_chan, outsize) -> None:
        super().__init__()
        self.hidden_dim = hidden_dim
        self.map = nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.ReLU(True))
        self.up = nn.Sequential(
            nn.ConvTranspose2d(hidden_dim, out_chan, outsize, 1, 0, bias=False),
            nn.BatchNorm2d(out_chan),
            nn.ReLU(True),
            nn.Conv2d(out_chan, out_chan, kernel_size=3, padding=1, stride=1),
            nn.BatchNorm2d(out_chan),
            nn.ReLU(True),
        )

    def forward(self, z):
        z = self.map(z)
        z = self.up(z.reshape(-1, self.hidden_dim, 1, 1))
        return z


class ResnetBlock(nn.Module):
    """
    Adapted from Taming transformers Repo
    https://github.com/CompVis/taming-transformers/blob/3ba01b241669f5ade541ce990f7650a3b8f65318/taming/modules/diffusionmodules/model.py
    """

    def __init__(self, *, in_channels, out_channels=None):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels

        self.norm1 = nn.GroupNorm(num_groups=32, num_channels=in_channels, eps=1e-6, affine=True)
        self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1)

        self.norm2 = nn.GroupNorm(num_groups=32, num_channels=out_channels, eps=1e-6, affine=True)
        self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
        if self.in_channels != self.out_channels:
            self.nin_shortcut = torch.nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, x):
        h = x
        h = self.norm1(h)
        h = h * torch.sigmoid(h)
        h = self.conv1(h)

        h = self.norm2(h)
        h = h * torch.sigmoid(h)
        h = self.conv2(h)

        if self.in_channels != self.out_channels:
            x = self.nin_shortcut(x)

        return x + h


class AttnBlock(nn.Module):
    """
    Adapted from Taming transformers Repo
    https://github.com/CompVis/taming-transformers/blob/3ba01b241669f5ade541ce990f7650a3b8f65318/taming/modules/diffusionmodules/model.py
    """

    def __init__(self, in_channels):
        super().__init__()
        self.in_channels = in_channels

        self.norm = nn.GroupNorm(num_groups=32, num_channels=in_channels, eps=1e-6, affine=True)
        self.q = torch.nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.k = torch.nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.v = torch.nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)
        self.proj_out = torch.nn.Conv2d(in_channels, in_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, x):
        h_ = x
        h_ = self.norm(h_)
        q = self.q(h_)
        k = self.k(h_)
        v = self.v(h_)

        # compute attention
        b, c, h, w = q.shape
        q = q.reshape(b, c, h * w)
        q = q.permute(0, 2, 1)  # b,hw,c
        k = k.reshape(b, c, h * w)  # b,c,hw
        w_ = torch.bmm(q, k)  # b,hw,hw    w[b,i,j]=sum_c q[b,i,c]k[b,c,j]
        w_ = w_ * (int(c) ** (-0.5))
        w_ = torch.nn.functional.softmax(w_, dim=2)

        # attend to values
        v = v.reshape(b, c, h * w)
        w_ = w_.permute(0, 2, 1)  # b,hw,hw (first hw of k, second of q)
        h_ = torch.bmm(v, w_)  # b, c,hw (hw of q) h_[b,c,j] = sum_i v[b,c,i] w_[b,i,j]
        h_ = h_.reshape(b, c, h, w)

        h_ = self.proj_out(h_)

        return x + h_


class MTLDecoder(SharedBlock):
    """
    Builds two-dimensional data-domain samples given
    """

    class Config(SharedBlock.Config):
        module_name: str = "mtldecoder"
        pixel_embedding_dim: int = 16
        out_channels: int = 3
        latent_upsamp_size: int = 8
        use_pyramid_channels: List[bool] = [True, False, False, False]
        include_latent_representation: bool = False
        num_residual_blocks: int = 3

    def __init__(
        self,
        args: Config,
        enc_out_channels: List[int],
        enc_strides: List[int],
        enc_hiddendim: int,
    ):
        super().__init__(args)
        self.args: MTLDecoder.Config
        # assert enc_strides[-1] == 32
        # assert enc_strides[1:] == [2**(i+2) for i in range(len(enc_out_channels[1:]))]
        # assert len(args.use_pyramid_channels) == len(enc_out_channels) - 1, "Wrong number of masked"

        # remove first skip with same spatial resolution
        decoder_channels = [(2**i) * args.pixel_embedding_dim for i in reversed(range(len(enc_out_channels) - 1))]
        encoder_channels = enc_out_channels[1:]
        # reverse channels to start from head of encoder
        encoder_channels = [x if args.use_pyramid_channels[i] else 0 for i, x in enumerate(encoder_channels[::-1])]

        # computing blocks input and output channels
        in_channels = [encoder_channels[0]] + list(decoder_channels[:-1])
        if self.args.include_latent_representation:
            in_channels[0] += enc_hiddendim
        skip_channels = [c for c in encoder_channels[1:]] + [0]
        blocks = [
            DecoderBlock(in_ch, skip_ch, out_ch)
            for in_ch, skip_ch, out_ch in zip(in_channels, skip_channels, decoder_channels)
        ]
        if self.args.include_latent_representation:
            self.upsamp = LatentBlock(enc_hiddendim, enc_hiddendim, self.args.latent_upsamp_size)

        blocks[0] = DecoderBlock(in_channels[0], skip_channels[0], decoder_channels[0], attention_type="scse")
        blocks.append(DecoderBlock(args.pixel_embedding_dim, 0, args.pixel_embedding_dim))
        # for b in blocks:
        #     [c.apply(weights_init) for c in b.children() if isinstance(c, nn.Conv2d)]
        self.blocks = nn.ModuleList(blocks)

        if self.args.num_residual_blocks > 0:
            resblocks = []
            # in-conv
            resblocks.append(nn.Conv2d(in_channels[0], in_channels[0], kernel_size=3, padding=1, stride=1))
            for _ in range(self.args.num_residual_blocks):
                resblocks.append(ResnetBlock(in_channels=in_channels[0], out_channels=in_channels[0]))
                resblocks.append(AttnBlock(in_channels=in_channels[0]))

            # norm out
            resblocks.append(nn.GroupNorm(num_groups=32, num_channels=in_channels[0], eps=1e-6, affine=True))
            # out-conv
            resblocks.append(nn.Conv2d(in_channels[0], in_channels[0], kernel_size=3, padding=1, stride=1))
            self.residuals = nn.ModuleList(resblocks)

        self.make_mtl_compatible()
        # self.apply(weights_init)

    def get_output_dim_per_pixel(self) -> int:
        return self.args.pixel_embedding_dim

    def get_upsampling_factor(self) -> int:
        return 1

    def forward(self, features: List[torch.Tensor], z: torch.Tensor):
        """
        Computes the pixel embedding. Useful for the forward pass of segmentation tasks.
        """
        # remove first skip with same spatial resolution (which is the original input)
        features = features[1:]
        # reverse channels to start from head of encoder ()
        features = features[::-1]

        skips = features[1:]

        # Build initial state of the upsampling routine
        if self.args.include_latent_representation:
            z_featurized = self.upsamp(z)
            z_featurized = F.interpolate(z_featurized, size=features[0].shape[2:], mode="nearest")
            x = z_featurized
            if self.args.use_pyramid_channels[0]:
                x = torch.concat([features[0], z_featurized], dim=1)
        else:
            x = features[0]

        if self.args.num_residual_blocks > 0:
            for i, resblock in enumerate(self.residuals):
                x = resblock(x)
        for i, decoder_block in enumerate(self.blocks):
            skip = skips[i] if i < len(skips) and self.args.use_pyramid_channels[i + 1] else None
            x = decoder_block(x, skip)

        return x



from pathlib import Path
from typing import List, Mapping, Sequence, cast, Union
from typing_extensions import Annotated

from pydantic import Field
import torch
import torch.nn as nn
import logging
import wandb

from mmm.neural.model_protocols import EncoderModel
from .SharedBlock import SharedBlock
from mmm.neural.activations import ActivationFn, ActivationFunctionConfig
from mmm.neural.pooling import GlobalPooling, GlobalPoolingConfig
from mmm.neural.modules.simple_cnn import MiniConvNet
from mmm.neural.modules.TorchVisionCNN import TorchVisionCNN
from mmm.neural.modules.swinformer import TorchVisionSwinformer
from mmm.neural.modules.TimmEncoder import TimmEncoder

EncoderArchitectureType = Union[
    MiniConvNet.Config,
    TorchVisionCNN.Config,
    TorchVisionSwinformer.Config,
    TimmEncoder.Config,
]


class PyramidEncoder(SharedBlock):
    class Config(SharedBlock.Config):
        model: Annotated[EncoderArchitectureType, Field(discriminator="architecture")] = TorchVisionCNN.Config()
        module_name: str = "encoder"
        pyramid_channels: dict[int, int] = Field(
            default={},
            description="Channels of the feature pyramid, can be left {} if the model's channels are already ok.",
        )

    def __init__(self, args: Config):
        super().__init__(args)
        self.args: PyramidEncoder.Config
        self.model: nn.Module = self.args.model.build_instance(self.args.pyramid_channels)

        self.remapper = nn.ModuleDict(
            {
                # Key of nn.ModuleDict cannot be an int
                f"{pyramid_index}": nn.Conv2d(
                    in_channels=self.model.get_feature_pyramid_channels()[pyramid_index],
                    out_channels=new_channels,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
                for pyramid_index, new_channels in self.args.pyramid_channels.items()
            }
        )

        self.make_mtl_compatible()

    def get_feature_pyramid_channels(self) -> List[int]:
        """
        Returns the channels of the input and all channels of the feature pyramid.
        For input [B, 2, 224, 224] and feature maps [B, 32, 128, 128], [B, 64, 96, 96] it would return [2, 32, 64]
        """
        channels_ls: list[int] = cast(EncoderModel, self.model).get_feature_pyramid_channels()
        for overwrite_channel, model_channel in self.args.pyramid_channels.items():
            channels_ls[overwrite_channel] = model_channel
        return channels_ls

    def get_strides(self) -> List[int]:
        """
        Indicates how much smaller each featuremap is than the input.
        For a feature-pyramid of input [B, C_input, 64, 64]:
        [B, C_input, 64, 64]
        [B, C_l1, 32, 32]
        [B, C_l2, 16, 16]

        The strides would be [1, 2, 4]
        """
        return cast(EncoderModel, self.model).get_strides()

    def forward(self, input: torch.Tensor) -> list[torch.Tensor]:
        feature_pyramid = self.model(input)

        for pyramid_index in self.remapper.keys():
            remap_conv = self.remapper[pyramid_index]
            feature_pyramid[int(pyramid_index)] = remap_conv(feature_pyramid[int(pyramid_index)])

        return feature_pyramid

    def get_example_input(self):
        return torch.rand(1, self.get_feature_pyramid_channels()[0], 224, 224).to(self.torch_device)

    def get_input_names(self) -> Sequence[str]:
        return ["input"]

    def get_output_names(self) -> Sequence[str]:
        return ["input_identity"] + [
            f"pyramid_{i}" for i in range(len(self.get_strides()) - 1)
        ]  # + ["image_embedding"]

    def get_dynamic_axes(self) -> Mapping[str, Mapping[int, str]]:
        return {
            "input": {0: "batch_size", 2: "height", 3: "width"},
            "input_identity": {0: "batch_size", 2: "height", 3: "width"},
            # "image_embedding": {0: "batch_size"},
            **{
                f"pyramid_{i}": {
                    0: "batch_size",
                    2: f"height_{stride}" if stride != 1 else "height",
                    3: f"width_{stride}" if stride != 1 else "width",
                }
                for i, stride in enumerate(self.get_strides()[1:])
            },
        }

    def export_to_onnx(self, path: Path, for_task: str = "") -> None:
        if self.args.model.architecture == "swinformer":
            raise NotImplementedError("ONNX export is not supported for Swinformer")
        return super().export_to_onnx(path, for_task)



"""
Avoid circular imports by not importing other mmm stuff here.
"""

from __future__ import annotations  # when types cannot be written from top to bottom
import cv2
import numpy as np
import torch
from pathlib import Path
from typing import Any, Literal, Generic, TypeVar

from pydantic import BaseModel, Field, model_validator, ConfigDict, field_validator

import torchvision.transforms.functional as TF
from torchvision.datasets.folder import default_loader

from .typing_utils import rgbnumpy_to_base64, get_colors


class MTLType(BaseModel):
    """
    Generic class for all MTL types
    """

    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_untyped(cls, legacy_obj):
        raise NotImplementedError()

    def to_untyped(self):
        raise NotImplementedError()

    def _repr_html_(self):
        return f"<p>{self.__class__.__name__} did not implement _repr_html_</p>"

    def _st_repr_(self) -> Any | None:
        """
        Default representation for streamlit
        """
        import streamlit as st

        # Uses the html representation _repr_html_ by default
        st.write(self)


class ImageAnnotation(MTLType):
    annotation_type: Literal["image_caption", "multiclass_segmentation", "multilabel_segmentation"]


class ImageCaption(ImageAnnotation):
    annotation_type: Literal["image_caption"] = "image_caption"
    caption: str = Field(..., description="The content of the image caption")

    @staticmethod
    def build_example():
        return ImageCaption(caption="A smiling female astronaut. She wears a space suit and holds a helmet.")


class MultiClassSegmentation(ImageAnnotation):
    annotation_type: Literal["multiclass_segmentation"] = "multiclass_segmentation"
    data: torch.Tensor = Field(..., description="The mask")

    @classmethod
    def from_untyped(cls, legacy_obj):
        return cls(data=legacy_obj)

    def to_untyped(self):
        return self.data

    @staticmethod
    def build_example(for_image: RGBImage):
        mask = torch.zeros(for_image.data.shape[1:]).long()
        mask[100:400, 100:400] = 1
        mask[200:300, 300:400] = 2
        return MultiClassSegmentation(data=mask)

    @field_validator("data")
    @classmethod
    def check_tensor(cls, v: torch.Tensor):
        assert isinstance(v, torch.LongTensor), "Mask should be a long tensor"
        assert len(v.shape) == 2, "Mask should be (H, W)"
        return v

    def _repr_html_(self):
        # Convert the mask to an RGB representation between 0 and 255.
        np_img = self.data.numpy()
        np_img = np_img.astype(np.float32) * (255 / np.max(np_img))
        # Add three channels
        np_img = np.stack([np_img] * 3, axis=-1)
        img, desc = (
            rgbnumpy_to_base64(np_img),
            f"type: {self.data.dtype}, shape: {self.data.shape}",
        )
        return f"""
        <figure style="text-align: center">
            <img src="{img}" alt="{desc}">
            <figcaption style="text-align: center">{desc}</figcaption>
        </figure>
        """


class MultiLabelSegmentation(ImageAnnotation):
    annotation_type: Literal["multilabel_segmentation"] = "multilabel_segmentation"
    data: torch.Tensor = Field(..., description="The mask")
    class_names: list[str] | None = None

    @classmethod
    def from_untyped(cls, legacy_obj):
        return cls(data=legacy_obj)

    def to_untyped(self):
        return self.data

    @staticmethod
    def build_example(for_image: RGBImage):
        mask = torch.zeros([4] + list(for_image.data.shape[1:])).long()
        # mask[0] is background
        mask[1, 100:400, 100:400] = 1
        mask[2, 200:300, 300:400] = 1
        mask[3, 100:400, 200:350] = 1
        # mask[3, 200:300, 300:400] = 1
        return MultiLabelSegmentation(data=mask, class_names=["background", "class1", "class2", "class3"])

    @field_validator("data")
    @classmethod
    def check_tensor(cls, v: torch.Tensor):
        assert isinstance(v, torch.LongTensor), "Mask should be a long tensor"
        assert len(v.shape) == 3, "Mask should be (#classnames, H, W)"
        return v

    def _repr_html_(self):
        np_mask = self.data.numpy()
        html = ""
        for class_i in range(self.data.shape[0]):
            # Convert the mask to an RGB representation between 0 and 255.
            np_img = np_mask[class_i]
            np_img = np_img.astype(np.float32) * (255 / np.max(np_img))
            # Add three channels
            np_img = np.stack([np_img] * 3, axis=-1)
            img, desc = (
                rgbnumpy_to_base64(np_img),
                f"type: {self.data.dtype}, shape: {self.data.shape}",
            )
            html += f"""
            <figure style="text-align: center">
                <img src="{img}" alt="{desc}">
                <figcaption style="text-align: center">{desc}</figcaption>
            </figure>
            """
        return html

    def st_class_selector(self, st_prefix: str) -> dict[str, bool]:
        import streamlit as st

        assert self.class_names is not None, "Class names should be set for selecting classes"

        colors = get_colors(len(self.class_names))
        res = {}
        for i, c in enumerate(self.class_names):
            indices, nums = torch.unique(self.data[i], return_counts=True)
            # Count 1s in the mask
            num_pixels = nums[indices == 1][0] if 1 in indices else 0
            # num_pixels = class_counts[i] if i in class_indices else 0

            if num_pixels > 0:
                colorrgb = f"{int(colors[i][0]*255)}, {int(colors[i][1]*255)}, {int(colors[i][2]*255)}"
                col1, col2 = st.columns(2)
                with col1:
                    show = st.checkbox(
                        c,
                        key=f"{st_prefix}check{c}",
                        value=(len(self.class_names) == 1) or (i >= 1),
                    )
                with col2:
                    st.markdown(
                        f"<span style='color: rgb({colorrgb})'>{num_pixels} pixels</span>",
                        unsafe_allow_html=True,
                    )
                res[i] = show
            else:
                res[i] = False
        return res

    def blend_with_image(
        self,
        im: np.ndarray,
        classes: dict[int, bool],
    ) -> None:
        im = (im.copy() / 255.0).astype(np.float32)
        mask = self.data.numpy().copy()
        assert len(mask.shape) == 3
        colors = get_colors(len(classes))

        for class_i, overlay_class in classes.items():
            if overlay_class:
                color_overlay = im.copy()
                color_overlay[mask[class_i] == 1] = np.asarray(colors[class_i], dtype=np.float32)
                im = cv2.addWeighted(im, 0.7, color_overlay.astype(np.float32), 0.3, 0)

        return (im * 255).astype(np.uint8)


class RGBImage(MTLType):
    """
    Generic image sample with 3 channels
    """

    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)
    data: torch.Tensor = Field(..., description="The image")

    @classmethod
    def from_path(cls, img_path: Path):
        return cls(data=TF.to_tensor(default_loader(img_path)))

    @staticmethod
    def build_example():
        from skimage.data import astronaut

        return RGBImage(data=TF.to_tensor(astronaut()))

    @field_validator("data")
    @classmethod
    def check_tensor(cls, v: torch.Tensor):
        assert isinstance(v, torch.FloatTensor), "Image should be a float tensor"
        assert v.shape[0] == 3, "Image should have 3 channels"
        assert len(v.shape) == 3, "Image should be (RGB, H, W)"
        return v

    @classmethod
    def from_untyped(cls, legacy_obj):
        return cls(data=legacy_obj["image"])

    def to_untyped(self):
        return self.data

    def to_numpy(self) -> np.ndarray:
        return self.data.numpy().transpose(1, 2, 0) * 255

    def get_description(self) -> str:
        res = f"type: {self.data.dtype}, shape: {self.data.shape}<br />"
        res += f"Min: {torch.min(self.data):.2f}, Max: {torch.max(self.data):.2f}"
        return res

    def _repr_html_(self):
        img, desc = rgbnumpy_to_base64(self.to_numpy()), self.get_description()
        return f"""
        <figure style="text-align: center">
            <img src="{img}" alt="{desc}">
            <figcaption style="text-align: center">{desc}</figcaption>
        </figure>
        """


# Typevar that can be any child of ImageAnnotation
ImageAnnotationType = TypeVar("ImageAnnotationType", bound=ImageAnnotation)


class AnnotatedImage(MTLType, Generic[ImageAnnotationType]):
    """
    Generic classification sample
    """

    image: RGBImage = Field(..., description="The image")
    # Use default_factory just to be sure with the mutable default
    annotations: list[ImageAnnotationType] = Field(default_factory=lambda: [], description="The annotations")

    original_size: tuple[int, int] | None = Field(None, description="The size to be used for metric computation")

    @model_validator(mode="after")
    def check_annotation_compatibility(self):
        for anno in self.annotations:
            if isinstance(anno, MultiClassSegmentation):
                assert self.image.data.shape[1:] == anno.data.shape, "Annotation shape should match image shape"
            elif isinstance(anno, MultiLabelSegmentation):
                assert self.image.data.shape[1:] == anno.data.shape[1:], "Annotation shape should match image shape"
            elif isinstance(anno, ImageCaption):
                assert anno.caption, "Caption should always contain something"
            else:
                raise NotImplementedError(f"Unknown annotation type {anno.annotation_type}")
        # self.mask.data.shape == self.image.data.shape[1:]

    @staticmethod
    def build_example():
        img = RGBImage.build_example()
        return AnnotatedImage(
            image=img,
            annotations=[
                MultiClassSegmentation.build_example(for_image=img),
                ImageCaption.build_example(),
            ],
        )

    @staticmethod
    def blend_with_multiclass_mask(
        im: np.ndarray,
        mask: np.ndarray,
        classes: dict[int, bool],
    ) -> None:
        # Writing annotations on top of the image without copying first can lead to errors
        im = im.copy().astype(np.float32) / 255.0
        assert len(mask.shape) == 2
        colors = get_colors(len(classes))
        color_overlay = im.copy()
        for class_i, overlay_class in classes.items():
            if overlay_class:
                color_overlay[mask == class_i] = np.asarray(colors[class_i], dtype=np.float32)
        im = (cv2.addWeighted(im, 0.7, color_overlay.astype(np.float32), 0.3, 0) * 255).astype(np.uint8)
        return im

    def build_repr(
        self,
        viewer: Literal["bare_html", "streamlit", "jupyter"] = "bare_html",
        st_prefix: str = "",
    ) -> str:
        """
        Reusable functionality for html-based visualization.
        """
        img = self.image.to_numpy().copy().astype(np.uint8)

        for mcmask in filter(lambda x: x.annotation_type == "multiclass_segmentation", self.annotations):
            mask_classes = {i: True for i in range(torch.max(mcmask.data).item() + 1)}
            mask_classes[0] = False
            img = self.blend_with_multiclass_mask(img, mcmask.data.numpy(), mask_classes)

        for i, mlmask in enumerate(
            filter(
                lambda x: x.annotation_type == "multilabel_segmentation",
                self.annotations,
            )
        ):
            mlmask: MultiLabelSegmentation
            if viewer == "streamlit":
                mask_classes = mlmask.st_class_selector(st_prefix=f"{st_prefix}mlmask_{i}")
            else:
                mask_classes = {i: True for i in range(mlmask.data.shape[0])}
                mask_classes[0] = False
            img = mlmask.blend_with_image(img, mask_classes)

        captions = [x.caption for x in self.annotations if x.annotation_type == "image_caption"]

        desc = self.image.get_description()

        img = f"""
        <figure style="text-align: center">
            <img src="{rgbnumpy_to_base64(img)}" style="width: 50%">
            <figcaption style="text-align: center">{desc}</figcaption>
        </figure>
        """
        return img + "".join([f'<p style="text-align: center">Caption: {x}</p>' for x in captions])

    def _repr_html_(self, st_prefix: str = ""):
        return self.build_repr(viewer="bare_html", st_prefix=st_prefix)

    def _st_repr_(self, st_prefix: str = "") -> Any | None:
        return self.build_repr(viewer="streamlit", st_prefix=st_prefix)

    @classmethod
    def from_untyped(cls, legacy_obj: dict[str, Any]) -> AnnotatedImage:
        annotations = []

        if "captions" in legacy_obj:
            annotations.extend([ImageCaption(caption=text) for text in legacy_obj["captions"]])

        return cls(
            image=RGBImage.from_untyped({"image": legacy_obj["image"]}),
            annotations=annotations,
        )


__all__ = [
    "RGBImage",
    "ImageCaption",
    "MultiClassSegmentation",
    "AnnotatedImage",
    "MetaInfo",
]



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from pathlib import Path
from pydantic import Field
from pydantic_settings import BaseSettings
from urllib.parse import urljoin

try:
    from fastapi import FastAPI
    from fastapi.middleware.cors import CORSMiddleware

    from label_studio_sdk import Client
except ImportError:
    if not TYPE_CHECKING:
        FastAPI, CORSMiddleware, Client = Any, Any, Any

from mmm.data_loading.DistributedPath import DistributedPath
from mmm.labelstudio_ext.DLModel import DLModel
from mmm.labelstudio_ext.LGBModel import LGBModel
from mmm.labelstudio_ext.NativeBlocks import NativeBlocks, MMM_MODELS, DEFAULT_MODEL
from mmm.labelstudio_ext.LabelstudioCredentials import LabelstudioCredentials


class APISettings(BaseSettings):
    class Config:
        env_prefix = "MTLAPI_"

    modules_path: DistributedPath | str = MMM_MODELS[DEFAULT_MODEL]
    labelstudio_base: str = "http://datanodefec:9505"
    labelstudio_token: str = "1234567890"
    app_base: str = Field(
        "http://localhost:8000",
        description="The base URL of this service",
    )
    device_identifier: str = "cuda:0"
    dlconfig: DLModel.Config = DLModel.Config()
    lgbconfig: LGBModel.Config = LGBModel.Config()

    allow_cors: bool = True


def build_app(settings: APISettings) -> FastAPI:
    app = FastAPI()

    if settings.allow_cors:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
    app.settings = settings

    creds = LabelstudioCredentials(url=settings.labelstudio_base, token=settings.labelstudio_token)
    ls_client: Client = creds.build_client()

    nativeblocks = NativeBlocks(settings.modules_path, settings.device_identifier)
    app.model = nativeblocks

    dlmodel = DLModel(
        settings.dlconfig,
        ls_client,
        urljoin(settings.app_base, "/dl"),
        nativeblocks,
    )
    app.include_router(dlmodel.build_router(), prefix="/dl")

    lgbmmodel = LGBModel(
        settings.lgbconfig,
        ls_client,
        urljoin(settings.app_base, "/lgbm"),
        nativeblocks,
    )
    app.include_router(lgbmmodel.build_router(), prefix="/lgbm")

    @app.get("/status")
    async def status():
        return {
            "client_status": ls_client.check_connection(),
            "modules_path": settings.modules_path,
        }

    return app


if __name__ == "__main__":
    import uvicorn

    app = build_app(APISettings())

    uvicorn.run("mmm.inference_api:app", host="0.0.0.0", port=8000, reload=True)



from __future__ import annotations
import os
from copy import deepcopy, copy
import traceback
import wandb
import json
import logging
from pathlib import Path
from pydantic import Field

from typing_extensions import Annotated
from typing import Any, Callable, Dict, List, Optional, Iterable, Tuple, Literal

import itertools

import torch
import torch.nn as nn
import torch.multiprocessing as mp

from mmm.mtl_modules.MTLModule import MTLModule
from mmm.optimization.MTLOptimizer import MTLOptimizer
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.tasks.TaskModule import TaskModule
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.trainer.Loop import Loop
from mmm.utils import remove_folder_blocking_if_exists, recursive_equality
from mmm.trainer.TaskPurpose import TaskPurpose
from mmm.DataSplit import DataSplit
from .CallbackType import CallbackType
from mmm.event_selectors import (
    EventSelector,
    FixedEventSelector,
    RecurringEventSelector,
)
from mmm.BaseModel import BaseModel
from mmm.data_loading.DistributedPath import DistributedPath
from mmm.optimization.MTLOptimizer import MTLOptimizer
from mmm.trainer.Loop import TrainLoopConfig, ValLoopConfig

import torch.distributed as dist
from torch.nn.parallel.distributed import DistributedDataParallel


class EarlyStoppingConfig(BaseModel):
    """
    Groups setting which concern early stopping.
    """

    early_stopping_patience: int = Field(
        default=10,
        description="Number of downstream training loops with no improvement after which training will be stopped. "
        + "Criterion checking for early stopping starts after 2*patience loops.",
    )
    improvement_delta: float = Field(
        default=0.03,
        description="""
Minimum loss improvement to count as an improvement for the patience attribute of early stopping.
If early stopping is used, the trainer saves a checkpoint "earlystop" after each best loop.
""",
    )
    min_train_loops: int = 5
    criterion: Literal["trainloss", "valloss"] = "trainloss"


class MTLTrainer:
    class Config(BaseModel):
        train_device: Literal["cpu", "cuda"] = "cuda"
        max_epochs: int = Field(
            default=150,
            description="Maximum number of epochs. Can be overriden by fit().",
        )

        checkpoint_cache_folder: Path = Field(
            default_factory=lambda: Path(os.getenv("ML_DATA_OUTPUT", default="./")) / "trainer_checkpoints",
            description="""
The parent folder where the trainer will store checkpoints.
            """,
        )

        optim: MTLOptimizer.Config = Field(
            default=MTLOptimizer.Config(),
            description="If set, takes precedence over all tasks and shared block's optimizers.",
        )

        mtl_train_selector: Annotated[EventSelector, Field(discriminator="selector_type")] = RecurringEventSelector(
            every_n=1, starting_at=1
        )
        mtl_train_loop: TrainLoopConfig = TrainLoopConfig()

        mtl_validation_selector: Annotated[
            EventSelector, Field(discriminator="selector_type")
        ] = RecurringEventSelector(every_n=1)
        mtl_val_loop: ValLoopConfig = ValLoopConfig()

        early_stopping: EarlyStoppingConfig | None = None

        keep_checkpoints_of_epochs: Annotated[EventSelector, Field(discriminator="selector_type")] = FixedEventSelector(
            at_iterations=[]
        )

        rebuild_dataloaders: Annotated[EventSelector, Field(discriminator="selector_type")] = FixedEventSelector(
            at_iterations=[0]
        )

    class State(BaseModel):
        epoch: int
        step_counters: dict[str, int]
        best_val_losses: list[float] | None = None
        avg_losses_per_epoch: list[tuple[float | None, float | None]] = []
        stage_index: int = 0

    def __init__(
        self,
        args: Config,
        experiment_name: str,
        global_rank: int = int(os.getenv("RANK", default=0)),
        local_rank: int = int(os.getenv("LOCAL_RANK", default=0)),
        local_world_size: int = int(os.getenv("LOCAL_WORLD_SIZE", default=1)),
        world_size: int = int(os.getenv("WORLD_SIZE", default=1)),
        clear_checkpoints: bool = False,
    ):
        self.experiment_name = experiment_name
        self.rank, self.world_size = global_rank, world_size
        self.local_rank, self.local_world_size = local_rank, local_world_size
        self.mtl_tasks: List[MTLTask] = []
        # self.eval_tasks: List[MTLTask] = []
        self.args: MTLTrainer.Config = args
        self.stages = [""]

        # Not a nn.ModuleDict, because the trainer is not an nn.Module
        self.shared_blocks: Dict[str, SharedBlock] = {}
        self.ddp_model = None

        self.state = self.State(epoch=0, step_counters={}, best_val_losses=None, avg_losses_per_epoch=[])

        self.experiment_ckpt_folder = self.args.checkpoint_cache_folder / self.experiment_name
        if self.rank == 0:
            if clear_checkpoints:
                remove_folder_blocking_if_exists(self.experiment_ckpt_folder)
            if not self.experiment_ckpt_folder.exists():
                self.experiment_ckpt_folder.mkdir(parents=True)
                logging.info(f"Created trainer's cache folder: {self.experiment_ckpt_folder}")
        if dist.is_initialized():
            dist.barrier()

        self.mtl_optimizer = None
        self.callbacks: Dict[CallbackType, Dict[str, Callable]] = {cb_type: {} for cb_type in CallbackType}

    def __repr__(self) -> str:
        res = f"Trainer config: {self.args.model_dump()}\n"
        res += f"Trainer state: {self.state}\n"
        res += f"{self.rank=}, {self.local_rank=}, {self.world_size=}, {self.local_world_size=}\n"
        res += f"Tasks used for pretraining:{[t.get_name() for t in self.mtl_tasks]}\n"
        return res

    def get_shared_block(self, n: str) -> SharedBlock:
        return self.shared_blocks[n]

    def add_shared_block(self, v: SharedBlock) -> None:
        if torch.cuda.is_available():
            self.shared_blocks[v.args.module_name] = v.set_device(self.args.train_device)
        else:
            logging.warning("CUDA is not available, using CPU instead.")
            self.shared_blocks[v.args.module_name] = v

    def add_shared_blocks(self, vs: Iterable[SharedBlock]) -> MTLTrainer:
        for shared_block in vs:
            self.add_shared_block(shared_block)
        return self

    def add_mtl_task(self, t: MTLTask):
        for task in self.mtl_tasks:
            if task.get_name() == t.get_name():
                raise Exception(f"{t.get_name()} is already known to the trainer.")

        if torch.cuda.is_available():
            self.mtl_tasks.append(t.set_device(self.args.train_device))
        else:
            logging.warning("CUDA is not available, using CPU instead.")
            self.mtl_tasks.append(t)

        for block in self.shared_blocks.values():
            block.prepare_for_task(t.get_name())

        return t

    def get_task_by_name(self, task_name: str) -> MTLTask | None:
        for t in self.mtl_tasks:
            if t.get_name() == task_name:
                return t
        return None

    def callback_after_each_epoch(self, shared_model_train_mode=False, use_inference_mode=True):
        """
        The trainer will call the decorated function each epoch.
        """

        # Python calls this with the function `f` of the user when defining `f`
        def setup_callback(cb: Callable[[MTLTrainer], None]) -> Callable[[MTLTrainer], None]:
            logging.info(f"Trainer will call {cb.__name__} after each epoch.")
            logging.info(f"It will apply {shared_model_train_mode=} to all MTLModules")
            logging.info(f"It will set the default task for all task-specific parts of shared blocks")
            logging.info(f"If you want to use the task-specific parts of a task, use `set_active_task(...)`")
            logging.info("It will wrap your callback in torch's inference mode:")

            def wrapper(trainer: MTLTrainer) -> None:
                for m in self.mtl_tasks:  # + self.eval_tasks:  # type: ignore
                    m.prepare_epoch(
                        self.state.epoch,
                        f"cb_{cb.__name__}",
                        training_mode=shared_model_train_mode,
                    )
                if self.ddp_model is not None:
                    self.ddp_model.train(shared_model_train_mode)

                for shared_block in self.shared_blocks.values():
                    shared_block.set_active_task("original")

                with torch.inference_mode(mode=use_inference_mode):
                    return cb(trainer)

            self.callbacks[CallbackType.each_epoch][cb.__name__] = wrapper
            return wrapper

        return setup_callback

    def get_task_module(self, task_name: str) -> TaskModule:
        """
        Builds a task module for the given task name, wrapping its dependencies from the shared blocks.

        If the shared blocks have task-specific parts, they will be set to the exported task.
        """
        task = self.get_task_by_name(task_name)
        if task is None:
            raise Exception(f"Task {task_name} is not known to the trainer.")
        export_module = TaskModule(task, self.shared_blocks)
        for shared_module in export_module.shared_modules.values():  # type: ignore
            shared_module: SharedBlock
            shared_module.set_active_task(task.get_name())
        return export_module

    @torch.no_grad()
    def save_blocks_native(self, file_path: DistributedPath, only_inference: bool = True):
        """
        Wraps all shared blocks and tasks into a dictionary.

        If only_inference is True, the model is set to evaluation mode and the cohort objects are removed.

        This export is suitable for machine to machine exports that happen quickly and on the same codebase.
        If the mmm code is slightly changed, the exported file will break.
        In that case, use the ONNX export.
        """
        from mmm.labelstudio_ext.NativeBlocks import NativeBlocks

        exportdict = nn.ModuleDict(self.shared_blocks)

        for exporttask in self.mtl_tasks:
            if only_inference:
                # A shallow copy is enough, because we do not modify the deeper structures
                exporttask = copy(exporttask)
                exporttask.cohort = None
            exportdict[exporttask.get_name()] = exporttask
        if only_inference:
            exportdict = exportdict.eval().cpu()
        NativeBlocks.save_to_disk(file_path, exportdict)
        return exportdict

    @torch.no_grad()
    def save_task_native(
        self,
        task_name: str,
        file_path: Path | None = None,
        only_inference: bool = True,
    ):
        """
        If file_path is none, will save to ./task_name.pt

        The model can be used like this:

        ```
        # Load the model using PyTorch native saving
        export_task_id = "taskname"
        model = torch.load(f"{export_task_id}.pt")
        model = model.eval()

        with torch.inference_mode():
            model(inputbatch: torch.Tensor)
        ```

        Disadvantage of this (from PyTorch docs):

        The disadvantage of this approach is that the serialized data is bound to the specific classes
        and the exact directory structure used when the model is saved.
        The reason for this is because pickle does not save the model class itself.
        Rather, it saves a path to the file containing the class, which is used during load time.
        Because of this, your code can break in various ways when used in other projects or after refactors.
        """
        if file_path is None:
            file_path = Path(f"./{task_name}.pt")

        export_module = self.get_task_module(task_name)

        try:
            # Workers cannot be pickled, kill them!
            export_module.task.cohort.terminate_workers()
        except Exception as e:
            logging.warning(f"Could not terminate workers: {e}")

        if only_inference:
            export_module.task = deepcopy(export_module.task)
            del export_module.task.cohort

        export_module.eval()

        torch.save(export_module, file_path)

    @torch.no_grad()
    def save_task_as_onnx(self, task_name: str, file_path: Path | None = None):
        """
        Given a task name (either a pretraining task or a downstream task), creates an onnx network.
        """
        if file_path is None:
            file_path = Path("./somenetwork.onnx")

        # ONNX import only relevant for this function, do not put at top to reduce startup time
        import torch.onnx

        task = self.get_task_by_name(task_name)
        assert task is not None
        onnx_net = TaskModule(task, self.shared_blocks)
        for shared_module in onnx_net.shared_modules.values():  # type: ignore
            shared_module: SharedBlock
            shared_module.set_active_task(task.get_name())
        example_x = onnx_net.task.cohort.get_onnx_input(f"cuda:{self.rank}")
        onnx_net.eval()

        torch.onnx.export(
            onnx_net,  # model being run
            example_x,  # model input (or a tuple for multiple inputs)
            str(file_path),  # where to save the model (can be a file or file-like object)
            export_params=True,  # store the trained parameter weights inside the model file
            opset_version=13,  # the ONNX version to export the model to
            do_constant_folding=True,  # whether to execute constant folding for optimization
            input_names=["input"],  # the model's input names
            output_names=["output"],  # the model's output names
            training=torch.onnx.TrainingMode.EVAL,
            dynamic_axes={
                "input": {0: "batch_size"},  # variable length axes
                "output": {0: "batch_size"},
            },
        )

    def create_checkpoint(self, prefix: str) -> Path:
        folder_path = self.experiment_ckpt_folder / f"{prefix}-{self.state.epoch}"
        if self.rank == 0:
            if folder_path.exists():
                logging.warning(f"Checkpoint {folder_path} already exists. Probably the writing was interrupted.")
            folder_path.mkdir(parents=True, exist_ok=True)

        # Make all processes wait until rank 0 has created the folder
        if dist.is_initialized():
            dist.barrier()

        logging.info(f"Saving checkpoint to {folder_path}")
        for k, v in self.shared_blocks.items():
            if self.rank == 0:
                v.save_checkpoint(folder_path / k)
            else:
                v.save_checkpoint(folder_path / f"{k}{self.rank}")

        # All checkpoints are created, now sanity check that all shared weights are identical
        if dist.is_initialized():
            dist.barrier()
            if self.rank == 0 and self.world_size > 1:
                for k, v in self.shared_blocks.items():
                    shouldbestate = str(
                        torch.load(
                            folder_path / k / "module.ckpt",
                            map_location=torch.device("cpu"),
                        )["model_state"]
                    )
                    for other_rank in range(0, dist.get_world_size()):
                        # Load the other state dict to cpu
                        pk = f"{k}{other_rank}" if other_rank != 0 else k
                        other_state = torch.load(
                            folder_path / pk / "module.ckpt",
                            map_location=torch.device("cpu"),
                        )
                        # if not recursive_equality(other_state, v.state_dict(), approx=True):
                        otherstate_str = str(other_state["model_state"])
                        if not otherstate_str == shouldbestate:
                            # Save both string to text files for debugging
                            with open(f"state_dict_{k}_rank0.txt", "w") as f:
                                f.write(shouldbestate)
                            with open(f"state_dict_{k}_rank{other_rank}.txt", "w") as f:
                                f.write(otherstate_str)
                            logging.error(f"Shared block {k} is not identical on rank 0 and rank {other_rank}")
                        else:
                            logging.debug(f"Shared block {k} is identical on rank 0 and rank {other_rank}")

        # Save the optimizer of rank 0
        if self.mtl_optimizer is None:
            logging.warning(f"Optimizer not initialized, initializing optimizer for: {folder_path}")
            self.init_optimizer()

        mtl_optim_folder = folder_path / "trainer_optim"
        self.mtl_optimizer.create_checkpoint(mtl_optim_folder, self.rank)

        # Always save the checkpoints of tasks assuming that different ranks each have unique tasks
        for task in self.mtl_tasks:
            task.save_checkpoint(folder_path / task.get_name())

        # Save the state last, this is the criterion to determine if a checkpoint is fully written
        if self.rank == 0:
            (folder_path / "meta.json").write_text(self.state.model_dump_json(indent=4))
        if dist.is_initialized():
            dist.barrier()
        return folder_path

    @staticmethod
    def verify_checkpoint(folder_path: Path) -> bool:
        try:
            with open(folder_path / "meta.json", "r+") as f:
                meta_dict = json.load(f)
                teststate = MTLTrainer.State(**meta_dict)
            return True
        except Exception as e:
            logging.info(f"Checkpoint {folder_path} is invalid due to {e}")
            return False

    def load_checkpoint(
        self,
        folder_path: Path,
        load_optim_state=True,
        load_meta=True,
        load_tasks=True,
    ) -> bool:
        """
        Uses an absolute folder path to enable loading checkpoints from other experiments.
        """
        if folder_path.exists():
            # First verify that the checkpoint is valid, we assume it is valid when the last thing was written:
            assert MTLTrainer.verify_checkpoint(folder_path), f"Checkpoint {folder_path} is invalid."

            shared_only_success = True
            for shared_module_name, shared_module in self.shared_blocks.items():
                try:
                    shared_module.load_checkpoint(folder_path / shared_module_name)
                    logging.info(f"Successfully loaded checkpoint of {shared_module_name}")
                except Exception as e:
                    logging.warning(f"Couldn't load checkpoint for shared block {shared_module_name} due to {e}")
                    shared_only_success = False

            if load_tasks:
                for task in self.mtl_tasks:
                    try:
                        if (folder_path / task.get_name()).exists():
                            task.load_checkpoint(folder_path / task.get_name())
                            logging.info(f"Successfully loaded checkpoint of {task.get_name()}")
                        else:
                            logging.warn(f"Couldn't find checkpoint of task {task.get_name()}")
                    except RuntimeError as e:
                        logging.warning(f"Couldn't load checkpoint for task {task.get_name()} due to {e}")
                    except ValueError as e:
                        logging.warning(f"Couldn't load checkpoint for task {task.get_name()} due to {e}")
                    except FileNotFoundError as e:
                        logging.warning(f"Couldn't find checkpoint for task {task.get_name()} due to {e}")

            if load_optim_state:
                try:
                    assert shared_only_success, "Cannot load optimizer state if there was a change to a shared block."
                    if self.mtl_optimizer is None:
                        self.init_optimizer()
                    self.mtl_optimizer.load_checkpoint(folder_path / "trainer_optim")
                except Exception as e:
                    logging.warn(f"Couldn't load optimizer checkpoint {folder_path / 'trainer_optim'} due to {e}")

            if load_meta:
                self.state = self.State(**json.loads((folder_path / "meta.json").read_text()))

            logging.info(f"Loaded checkpoint from {folder_path} with epoch {self.state.epoch}")
            return True
        else:
            logging.info(f"Checkpoint does not exist at {folder_path}")
            return False

    def _task_step(self, batch: Any, task: MTLTask):
        batch = task.prepare_batch(batch)
        return task.training_step(batch, self.ddp_model)

    def _controlsync(self, on: bool):
        self.ddp_model.require_backward_grad_sync = on  # type: ignore
        # for ddpmodel in self.ddp_model.values():
        #     ddpmodel.require_backward_grad_sync = on  # type: ignore

    def fix_worker_num(self, for_split: DataSplit) -> bool:
        allocated_workers = [
            sum([t.cohort.get_active_workers(split) for t in self.mtl_tasks])
            for split in [DataSplit.train, DataSplit.val]
        ]
        logging.info(f"{allocated_workers=} are already allocated for {for_split=}")
        take_workers_from = DataSplit.train if for_split is DataSplit.val else DataSplit.val
        if len(mp.active_children()) - allocated_workers[for_split.value] > len(os.sched_getaffinity(0)) // 2:
            logging.warning(f"Killing workers {take_workers_from=} because too many workers would be allocated")
            for t in self.mtl_tasks:
                t.cohort.terminate_datasplit_workers(take_workers_from)
            return True
        return False

    def run_mtl_train_epoch(self) -> Dict[MTLTask, List[float]]:
        """
        A pretrain epoch's purpose is training the shared blocks.

        Multi-task pre-training is a typical example.
        """
        self.fix_worker_num(for_split=DataSplit.train)

        # With distributed training this property of shared blocks is not autoforwarded to DDP wrapper
        self.ddp_model.train(True)

        loop = self.args.mtl_train_loop.build_instance(
            self.mtl_tasks,
            self.state.epoch,
            self._task_step,
            self.ddp_model,
            f"mtltrain{self.stages[self.state.stage_index]}",
            self.state.step_counters,
            self.mtl_optimizer,
            rank=self.rank,
            syncon_syncoff=(
                (
                    lambda: self._controlsync(True),
                    lambda: self._controlsync(False),
                )
                if dist.is_initialized()
                else None
            ),
        )
        return loop.drain_to_dict()

    def run_mtl_val_epoch(self) -> Dict[MTLTask, List[float]]:
        """Validates the pre-training tasks"""
        self.fix_worker_num(for_split=DataSplit.val)

        self.ddp_model.train(False)
        loop: Loop = self.args.mtl_val_loop.build_instance(
            self.mtl_tasks,
            self.state.epoch,
            self._task_step,
            self.ddp_model,
            f"mtlval{self.stages[self.state.stage_index]}",
            self.state.step_counters,
            None,  # No optimizer validation,
            rank=self.rank,
        )
        return loop.drain_to_dict()

    def check_workernum(self, auto_adjust=True):
        import math
        from numpy.random import choice
        import os

        def adjust(ls: List[int], max_num: int):
            ls = [math.ceil((x / sum(ls)) * max_num) for x in ls]

            # Now we should be really close but some tasks need to lose some workers. Take from the rich with a higher probab:
            while sum(ls) > max_num:
                i = choice(list(range(len(ls))), 1, p=[x / sum(ls) for x in ls])[0]
                ls[i] = max(0, ls[i] - 1)
            return ls

        workernum_available = len(os.sched_getaffinity(0)) // self.local_world_size

        # For training a safe number seems to be to use half the workers for pretraining and half for validation:
        workers = {task.args.module_name: task.cohort.args.num_workers for task in self.mtl_tasks}
        logging.info(f"{workernum_available} workers available for pretraining tasks. You requested {workers}")

        if auto_adjust and sum(workers.values()) > workernum_available:
            new_workers = adjust(list(workers.values()), workernum_available)
            for new_worker_num, task in zip(new_workers, self.mtl_tasks):
                task.cohort.args.num_workers = max(new_worker_num, 1)
            after_adjustment_workers = {task.args.module_name: task.cohort.args.num_workers for task in self.mtl_tasks}
            logging.info(f"Workers changed to {after_adjustment_workers}")
        else:
            assert sum(workers.values()) <= workernum_available, (
                f"You requested {workers} workers, "
                + f"but your environment should only use {workernum_available} workers in sum."
            )

    def prepare_dataloading(self) -> None:
        # In some environments, requesting more workers than there are CPU cores results in an error:
        self.check_workernum()

        for task in self.mtl_tasks:
            if self.args.rebuild_dataloaders.is_event(self.state.epoch):
                task.cohort.terminate_workers()
                task.cohort.data_loaders = (None, None)
            if None in task.cohort.data_loaders:
                task.cohort.prepare_epoch(epoch=self.state.epoch)

    def run_mtl_epoch(self) -> Tuple[Optional[List[float]], Optional[List[float]]]:
        """
        Returns None if no validation is performed. Otherwise, returns the average validation loss for each task.
        """
        self.prepare_dataloading()

        # Train the shared blocks and task heads:
        if self.args.mtl_train_selector.is_event(self.state.epoch):
            # Always unfreeze shared blocks for mtl epoch
            train_losses_by_task: Dict[MTLTask, List[float]] = self.run_mtl_train_epoch()
            train_result = [task_losses for _, task_losses in train_losses_by_task.items()]
        else:
            train_result = None

        if self.args.mtl_validation_selector.is_event(self.state.epoch):
            # Validate pre-training tasks
            val_losses_by_task: Dict[MTLTask, List[float]] = self.run_mtl_val_epoch()
            val_result: Optional[List[List[float]]] = [task_losses for _, task_losses in val_losses_by_task.items()]
        else:
            val_result = None

        if train_result is not None:
            self.mtl_optimizer.after_loop_step(self.state.epoch, train_result, val_result)
        else:
            logging.warn(f"Without training, no after_loop_step is performed ({self.state.epoch=})")

        train_avg = (
            [sum(task_losses) / len(task_losses) for task_losses in train_result] if train_result is not None else None
        )
        val_avg = (
            [sum(task_losses) / len(task_losses) for task_losses in val_result] if val_result is not None else None
        )
        if train_avg is not None:
            avg_train_loss = sum(train_avg) / len(train_avg)
        else:
            avg_train_loss = None
        if val_avg is not None:
            avg_val_loss = sum(val_avg) / len(val_avg)
        else:
            avg_val_loss = None
        self.state.avg_losses_per_epoch.append((avg_train_loss, avg_val_loss))
        return train_avg, val_avg

    def init_optimizer(self) -> None:
        self.ddp_model: SharedModules = SharedModules(self.shared_blocks)

        if dist.is_initialized():
            # find_unused_parameters is required whenever we have a shared block that is not used by all tasks
            self.ddp_model = DistributedDataParallel(
                self.ddp_model,
                device_ids=[int(self.local_rank)],
                find_unused_parameters=True,
            )  # type: ignore

        self.mtl_optimizer = MTLOptimizer(
            self.shared_blocks,
            self.args.optim,
            self.args.max_epochs,
            # Scale the gradients to account for DDP averaging gradients
            # While the docs say this is only relevant for the number of nodes:
            # "When a model is trained on M nodes with batch=N,
            # the gradient will be M times smaller
            # when compared to the same model trained on a single node with batch=M*N"
            # Other sources say that the gradient is averaged over all processes.
            gradient_scale_factor=float(self.world_size),
        )
        for task in self.mtl_tasks:
            self.mtl_optimizer.add_task(task)

    def get_summary_stats(self) -> Dict[str, Any]:
        module_summary = {}
        blocksummary: str = ""
        for blockname, block in self.shared_blocks.items():
            blocksummary = f"{blocksummary}<h2>{blockname}</h2>{block.__repr_html__()}"
        module_summary["sharedblocks"] = wandb.Html(blocksummary)
        for pretrain_module in self.mtl_tasks:
            module_summary[pretrain_module.get_name()] = wandb.Html(pretrain_module.__repr_html__())

        return module_summary

    def _fit(self, max_epochs: int):
        assert self.mtl_tasks, "No tasks added to trainer. Use `add_mtl_task` to add tasks."
        if self.state.epoch == 0:
            logging.info("Starting a new training. Logging module's stats")
            wandb.log({f"report/{k}": v for k, v in self.get_summary_stats().items()})
        elif self.state.epoch > max_epochs:
            logging.info(f"{max_epochs=} reached because {self.state.epoch=}")
            return
        else:
            logging.info(f"Resuming training at epoch {self.state.epoch=} until {max_epochs=}")

        for _ in itertools.count():
            self._run_callbacks(CallbackType.each_epoch, self)
            if self.rank == 0:
                self.cleanup_checkpoints()
            if dist.is_initialized():
                dist.barrier()
            with torch.no_grad():
                _, val_result = self.run_mtl_epoch()
                self.state.epoch += 1
                # This needs to be synchronized across processes to enable multi-GPU smart checkpointing

                if val_result is not None and (
                    self.state.best_val_losses is None or sum(val_result) < sum(self.state.best_val_losses)
                ):
                    self.state.best_val_losses = val_result  # list of average losses per task [t1loss, t2loss, ...]

                    if not dist.is_initialized():
                        self.create_checkpoint(f"bestbyvalidation{self.stages[self.state.stage_index]}")

            # Increment global epoch counter which is used for logging and is also checkpointed

            self.create_checkpoint("latest")

            if self.state.epoch > max_epochs:
                logging.info(f"Reached max_epochs ({max_epochs}). Stopping trainer.fit().")
                break

            if self.args.early_stopping is not None:
                recently_reported: List[Optional[float]] = [
                    t[0] if self.args.early_stopping.criterion == "trainloss" else t[1]
                    for t in self.state.avg_losses_per_epoch
                ]
                if recently_reported[-1] is None:
                    logging.warning(f"Epoch {self.state.epoch=} did not log a loss and is ignored in early stopping.")
                recent_losses: List[float] = list(filter(lambda x: x is not None, recently_reported))  # type: ignore
                if recent_losses:
                    if recently_reported[-1] is not None and min(recent_losses) == recent_losses[-1]:
                        self.create_checkpoint("earlystop")

                    if len(recent_losses) >= self.args.early_stopping.min_train_loops:
                        # Check if the last significant improvement was more than early_stopping_patience epochs ago
                        first_epoch_with_current_loss = min(
                            [
                                i
                                for i, l in enumerate(recent_losses)
                                if (l - self.args.early_stopping.improvement_delta) < recent_losses[-1]
                            ]
                        )
                        num_loops_without_improvement = len(recent_losses) - first_epoch_with_current_loss
                        if num_loops_without_improvement > self.args.early_stopping.early_stopping_patience:
                            logging.info(
                                f"Epoch {first_epoch_with_current_loss} already had a similar loss as {self.state.epoch=}. "
                                "Breaking training now. "
                                f"Relevant losses: {[f'{l:.2f}' for l in recent_losses]}"
                            )
                            break
                        else:
                            logging.info(
                                f"Relevant losses for early stopping: {[f'{l:.2f}' for l in recent_losses]}. "
                                f"Continuing because {num_loops_without_improvement=} "
                                f"is smaller than the patience of {self.args.early_stopping.early_stopping_patience}"
                            )

    def _run_callbacks(self, callback_type: CallbackType, *args, **kwargs):
        for cb_name, cb in self.callbacks[callback_type].items():
            logging.info(f"Running hook {cb_name} of type {callback_type}")
            cb(*args, **kwargs)

    def cleanup_checkpoints(self):
        """
        For every prefix, it only keeps checkpoints that are either the most recent of their prefix,
        or are marked to be saved in self.args.
        """
        checkpoints_and_states = [
            (p, self.State(**json.loads((p / "meta.json").read_text())))
            for p in self.experiment_ckpt_folder.iterdir()
            if MTLTrainer.verify_checkpoint(p)
        ]
        marked_for_removal = []
        for checkpoint_path, checkpointstate in checkpoints_and_states:
            if self.args.keep_checkpoints_of_epochs.is_event(checkpointstate.epoch):
                continue

            other_checkpoints_with_prefix = [
                (p, state)
                for p, state in checkpoints_and_states
                if p.name.startswith(checkpoint_path.name.split("-")[0])
            ]
            their_epochs = [state.epoch for _, state in other_checkpoints_with_prefix]
            if checkpointstate.epoch >= max(their_epochs):
                continue

            marked_for_removal.append(checkpoint_path)

        for remove_folder in marked_for_removal:
            logging.info(f"Removing checkpoint {remove_folder}")
            remove_folder_blocking_if_exists(remove_folder)
        return marked_for_removal

    @staticmethod
    def get_recent_checkpoint_paths(parent_folder: Path, prefix: str) -> List[tuple[Path, dict]]:
        # Determine the newest checkpoint and load from that
        checkpoints_and_states = [
            (p, MTLTrainer.State(**json.loads((p / "meta.json").read_text())))
            for p in parent_folder.iterdir()
            if MTLTrainer.verify_checkpoint(p) and p.name.startswith(prefix)
        ]
        return checkpoints_and_states

    def get_recent_checkpoint_path(self, prefix: str) -> Path | None:
        """
        Crawls through all checkpoints with that prefix and returns the Path of the checkpoint with the highest epoch.
        """
        checkpoints_and_states = self.get_recent_checkpoint_paths(self.experiment_ckpt_folder, prefix)
        if not checkpoints_and_states:
            return None
        checkpoint_epochs = [state.epoch for _, state in checkpoints_and_states]
        most_recent_index = checkpoint_epochs.index(max(checkpoint_epochs))
        return checkpoints_and_states[most_recent_index][0]

    def fit(self, until_epoch=None):
        """
        Runs a training. If not interactive it will create emergency checkpoints when an error occurs.

        If fit_for_epochs is set, it will finish if either args.max_epochs is reached or fit_for_epochs.
        """
        if self.mtl_optimizer is None:
            self.init_optimizer()
        if (latest_checkpoint_path := self.get_recent_checkpoint_path("latest")) is not None:
            self.load_checkpoint(
                latest_checkpoint_path,
                load_optim_state=True,
                load_meta=True,
                load_tasks=True,
            )
        return self._fit(until_epoch if until_epoch is not None else self.args.max_epochs)



from __future__ import annotations
import re
from typing import Callable, Dict, Optional, List, Literal, Tuple, TypeVar
from mmm.BaseModel import BaseModel
from pydantic import Field


class Bucket(BaseModel):
    expression: str
    description: Optional[str] = None
    expression_type: Literal["perfectmatch", "substring", "regex"] = "regex"

    def matches(self, s: str) -> bool:
        if self.expression_type == "perfectmatch":
            return s == self.expression
        elif self.expression_type == "substring":
            return self.expression in s
        elif self.expression_type == "regex":
            return bool(re.match(self.expression, s))
        else:
            raise Exception(f"Unknown expression type {self.expression_type}")

    def transform_to_class_name(self) -> str:
        if self.description:
            return self.description
        else:
            return self.expression

    def __str__(self) -> str:
        return self.transform_to_class_name()


class AmbiguousMatchException(Exception):
    pass


class BucketConfig(BaseModel):
    exactly_one_match: bool = Field(
        default=True,
        description="If true, only one bucket is allowed to match. Otherwise, the first match is returned.",
    )
    buckets: List[Bucket]

    def get_class_names(self) -> List[str]:
        descs = [b.transform_to_class_name() for b in self.buckets]
        assert len(set(descs)) == len(descs), f"Duplicate class names in {descs}"
        return descs

    def get_matches_for(self, desc: str) -> List[Bucket]:
        return [bucket for bucket in self.buckets if bucket.matches(desc)]

    def get_bucket(self, class_description: str) -> Tuple[Bucket, int]:
        matches = self.get_matches_for(class_description)
        assert len(matches) > 0, f"No match for {class_description}"

        if self.exactly_one_match and len(matches) != 1:
            raise AmbiguousMatchException(matches)

        return matches[0], self.buckets.index(matches[0])

    def __str__(self) -> str:
        return ", ".join([str(bucket) for bucket in self.buckets])

    def __eq__(self, other: BucketConfig):
        if not isinstance(other, BucketConfig):  # type: ignore
            return False

        return str(self) == str(other)

    def get_bucket_name(self, class_description: str) -> Tuple[str, int]:
        bucket, cls_id = self.get_bucket(class_description)
        return bucket.transform_to_class_name(), cls_id



import json
from functools import partial
import logging
from typing import Any
from torch.utils.data import Dataset
from mmm.BaseModel import BaseModel
from pathlib import Path

from .GeoAnno import GeoAnno
from .AnnoFilterConfig import AnnoFilterConfig


class WSIGeojsonDataset(Dataset[dict[str, Any]]):
    """
    Encapsulates WSIs and corresponding annotations.

    A WSI path can be anything, a local path, a URL, etc.
    """

    class Config(BaseModel):
        anno_filter: AnnoFilterConfig = AnnoFilterConfig()
        use_regions: AnnoFilterConfig | None = None
        ignore_regions: AnnoFilterConfig | None = None

    @staticmethod
    def load_annotations_from_json(json_path: Path) -> list[GeoAnno]:
        with open(json_path, "r") as f:
            json_dict = json.load(f)
            return [GeoAnno(x) for x in json_dict["features"]]

    def __init__(self, args: Config, wsi_anno_map: dict[Path, str]) -> None:
        self.args = args
        self.wsi_list = []

        outfiltered = []
        for anno_path, wsi_path in wsi_anno_map.items():
            annos: list[GeoAnno] = self.load_annotations_from_json(anno_path)

            original_anno_num = len(annos)
            # pre-filter annotations, e.g. by their class names

            if self.args.use_regions is not None:
                use_regions = list(filter(self.args.use_regions.apply, annos))
            else:
                use_regions = None

            if self.args.ignore_regions is not None:
                ignore_regions = list(filter(self.args.ignore_regions.apply, annos))
            else:
                ignore_regions = None
            anno_filter = partial(
                self.args.anno_filter.apply,
                use_regions=use_regions,
                ignore_regions=ignore_regions,
            )
            annos = list(filter(anno_filter, annos))

            # Only add this WSI if there are annotations left
            if annos:
                self.wsi_list.append((wsi_path, annos))
            else:
                outfiltered.append((wsi_path, original_anno_num))

        if outfiltered:
            logging.warning(f"Dropped {len(outfiltered)} WSI because no labels left after filtering: {outfiltered=}")

    def get_classes(self) -> list[str]:
        all_classes = set()
        for _, annos in self.wsi_list:
            for anno in filter(self.args.anno_filter.apply, annos):
                all_classes.add(anno.get_class_name())
        return list(all_classes)

    def __len__(self) -> int:
        return len(self.wsi_list)

    def __getitem__(self, index) -> dict[str, Any]:
        wsi_path, annos = self.wsi_list[index]

        return {"wsi": wsi_path, "annos": annos, "meta": {"id": str(wsi_path)}}

    def __repr__(self) -> str:
        wsi_stats = "".join([f"- {str(wsi_path)}: {len(annos)} annotations\n" for wsi_path, annos in self.wsi_list])
        return f"""
Length: {len(self)}\n\n
{wsi_stats}

{self.get_classes()}
        """



import logging
import os

LICENSE_URL = "https://github.com/FraunhoferMEVIS/MedicalMultitaskModeling/blob/main/license.md"
LICENSE_ENV_VAR_NAME = "MMM_LICENSE_ACCEPTED"
ACCEPTANCE_VALUE = "i accept"
ACCEPTANCE_PROMPT = f"Type {ACCEPTANCE_VALUE} to accept the MMM license agreement at {LICENSE_URL}:"
LICENSE_ACCEPT_INSTRUCTION = (
    f"Please accept the license agreement at {LICENSE_URL}"
    + f" by setting the environment variable {LICENSE_ENV_VAR_NAME} to '{ACCEPTANCE_VALUE}'."
)


def promptable():
    try:
        return get_ipython().__class__.__name__ == "ZMQInteractiveShell"  # type: ignore
    except:
        return False


def verify_license_accepted():
    if LICENSE_ENV_VAR_NAME not in os.environ or os.environ[LICENSE_ENV_VAR_NAME].lower() != ACCEPTANCE_VALUE:
        logging.error(LICENSE_ACCEPT_INSTRUCTION)

        if promptable():
            # Enable the user to accept the license agreement interactively
            user_input = input(ACCEPTANCE_PROMPT)
            if user_input.lower() != ACCEPTANCE_VALUE:
                raise Exception("You must accept the license agreement to proceed.")
            else:
                os.environ[LICENSE_ENV_VAR_NAME] = ACCEPTANCE_VALUE
            return verify_license_accepted()
        raise Exception(LICENSE_ACCEPT_INSTRUCTION)


verify_license_accepted()



from __future__ import annotations
import warnings
from enum import Enum, auto
import os
import numpy as np
import logging
from pathlib import Path
from abc import abstractmethod
import random
from typing import Callable, Any, List, Tuple, Optional, TypeVar, Set, NewType, Iterator
from typing import Generic, TypeVar
from torch.utils.data import Dataset


from tqdm.auto import tqdm
import torch
from torch.utils.data import Dataset, DataLoader, IterableDataset
from torch.utils.data._utils.collate import default_collate
from mmm.BaseModel import BaseModel
from mmm.transforms import KeepOnlyKeysInDict
from mmm.utils import disk_cacher, unique_str_hash, recursive_equality


class InvalidCaseError(Exception):
    def __init__(self, src_ds: MTLDataset, case_id: int, user_msgs: List[str]) -> None:
        self.src_ds, self.case_id = src_ds, case_id

        findcase = "Case" if case_id == -1 else f"Case {case_id}"
        s = f"{findcase} of {src_ds} is invalid:\n"
        for user_msg in user_msgs:
            s += f"{user_msg}\n"
        super().__init__(s)


SrcCaseType = dict[str, Any]


def mtl_collate(batch: List[SrcCaseType]) -> SrcCaseType:
    """Ignores the meta key when collating"""
    meta_dicts = [item["meta"] if "meta" in item else {} for item in batch]
    for case_dict in batch:
        if "meta" in case_dict:
            case_dict.pop("meta")
    res = default_collate(batch)
    res["meta"] = meta_dicts  # type: ignore
    return res  # type: ignore


def mtl_batch_collate(batch: List[List[SrcCaseType]]) -> SrcCaseType:
    """Ignores the meta key when collating datasets that already return batches"""
    # Flatten list
    batch_single_list = [item for sublist in batch for item in sublist]
    return mtl_collate(batch_single_list)


class DatasetStyle(Enum):
    MapStyle = auto()
    IterStyle = auto()


class IterableDSWrapper(IterableDataset):
    def __init__(self, fn) -> None:
        self.fn = fn

    def __iter__(self):
        return self.fn.__iter__()


class MTLDataset(Dataset[SrcCaseType]):
    """Wrapper for the PyTorch implementation of Datasets.
    Map-style datasets have a length and are integer-indexable.

    It enforces cases to be based on dictionaries where certain keys have a fixed place and meaning.
    Like "image" is usually the image data scaled to [0, 1].
    Custom types might make the code more modular but would break compatibility
    with MONAI transforms and PyTorch dataloaders.

    MTLDataset can be a wrapper around a vanilla torch dataset which can verify cases depending on the dataset type.
    Additionally, we recommend not using random transforms as part of the `src_ds` or the optional src_transform.

    Building steps for a case:

    1. Source dataset -> should contain the full data for a case
    2. src_transform -> converts the data into the right formats and the result should be small if cache_folder is set

    If cache_folder is set, the result after src_transform and stripping non-relevant keys is saved to `cache_folder`.
    Make sure `cache_folder` is unique to this dataset, including its config!
    If you are using our cohort abstraction it will try to make your cache unique w.r.t. its config.

    If you want to apply random transformations, use the cohort's batch-transforms
    """

    @staticmethod
    def assert_image_data_assumptions(image_data: torch.Tensor) -> None:
        assert torch.is_tensor(image_data) and type(image_data) is torch.Tensor, "image should be of type tensor"
        assert image_data.shape[0] >= 1, "Image needs to have at least one channel"
        assert isinstance(image_data, torch.FloatTensor)

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        mandatory_case_keys: List[str],
        optional_case_keys: List[str],
        src_transform: Optional[Callable[[SrcCaseType], SrcCaseType]] = None,
        batch_transform: Optional[Callable[[SrcCaseType], Any]] = None,
        collate_fn: Optional[Callable[[List[Any]], SrcCaseType]] = None,
        verify_cases: int = 0,
        **dataloader_kwargs,
    ) -> None:
        """Generates and prepares the dataset.

        Args:
            data_root (Path): Should be the path to a folder where the data for this dataset can be stored
            data_cache (Path): Path to cache directory. If it doesn't exist, the dataset will create it.
            mandatory_case_keys (List[str]): To be a valid case, these keys need to be in every case
            optional_case_keys: (List[str]): These keys will not be deleted from the case
            needs_preparation: (bool): If true, will try to call `self.prepare()`
        """
        self.src_ds = src_ds
        self.cache_folder: Optional[Path] = None
        self.src_transform = src_transform
        self.batch_transform: Optional[Callable] = batch_transform
        self.collate_fn = collate_fn if collate_fn is not None else mtl_collate
        self.dataloader_kwargs = dataloader_kwargs

        self.mandatory_case_keys: List[str] = mandatory_case_keys
        self.optional_case_keys: List[str] = optional_case_keys
        self.data_stripper = KeepOnlyKeysInDict(
            keys=set(self.mandatory_case_keys + self.optional_case_keys),
        )

        self.reduced_size: float = 1.0
        # Represented by a list because it is valid to repeat indices if the user wants so by setting a fraction > 1.
        if self.get_dataset_style() is DatasetStyle.MapStyle:
            self._indices: np.ndarray
            self.reset_indices()

            for case_id in random.sample(list(self._indices), k=min(len(self._indices), verify_cases)):
                self.verify_case_by_index(case_id)

    def get_dataset_style(self) -> DatasetStyle:
        if isinstance(self.src_ds, IterableDataset):
            return DatasetStyle.IterStyle
        else:
            return DatasetStyle.MapStyle

    def enable_caching(self, cache_folder: Path, test_cache_validity: bool = True):
        """
        Makes sure that the cache folder exists.

        If it already exists, optionally check for validity.
        """
        self.cache_folder = cache_folder
        if self.cache_folder.exists():
            if test_cache_validity:
                for case_path in self.cache_folder.glob("*.pkl"):
                    case_id: int = int(case_path.stem)
                    logging.info(f"Checking cache of case {case_id} for correctness")
                    shouldbe = self.get_untransformed_case(case_id)
                    cached_case = self.get_case_from_cache(case_id)
                    for key in self.mandatory_case_keys + self.optional_case_keys:
                        if key in shouldbe and key != "meta":
                            if not recursive_equality(shouldbe[key], cached_case[key]):
                                logging.warn(f"{key} shows difference between original and cached:")
                                logging.warn(f"{shouldbe[key]} and {cached_case[key]}")
                                assert False
                    break
        else:
            self.cache_folder.mkdir(parents=True)
            logging.info(f"Starting a new cache at {self.cache_folder}")
        return self

    def get_dataloader(self, **kwargs):
        if self.get_dataset_style() is DatasetStyle.MapStyle:
            if "sampler" in self.dataloader_kwargs and "shuffle" in kwargs:
                kwargs.pop("shuffle")
            return DataLoader(
                self,
                collate_fn=self.collate_fn,
                # a whole day in seconds
                # without multiprocessing, a timeout>0 is not allowed
                timeout=(86400 if "num_workers" in kwargs and kwargs["num_workers"] > 0 else 0),
                **kwargs,
                **self.dataloader_kwargs,
            )
        else:
            # Shuffle is not a valid option for iterablestyle datasets
            # assert isinstance(self, IterableDataset)
            if "shuffle" in kwargs:
                kwargs.pop("shuffle")
            return DataLoader(
                IterableDSWrapper(self),
                collate_fn=self.collate_fn,
                timeout=(86400 if "num_workers" in kwargs and kwargs["num_workers"] > 0 else 0),
                **kwargs,
                **self.dataloader_kwargs,
            )

    def get_mp_batchiterator(self, get_untransformed_cases=False, batch_size=1, shuffle=True, **kwargs):
        """
        Iterates through this dataset's batches with all available workers.
        """

        class UntransformedDS(Dataset):
            def __init__(self, src_ds: MTLDataset) -> None:
                self.src = src_ds

            def __len__(self) -> int:
                return self.src.__len__()

            def __getitem__(self, index) -> Any:
                return (
                    self.src.get_untransformed_case(index)
                    if self.src.cache_folder is None
                    else self.src.get_case_from_cache(index)
                )

        workernum_available = len(os.sched_getaffinity(0))
        # Use all workers for big datasets, use only few workers for small datasets
        used_workers = min(workernum_available, max(1, len(self) // (batch_size * 5)))
        dataloader_settings = dict(
            shuffle=shuffle,
            batch_size=batch_size,
            num_workers=used_workers,
            collate_fn=self.collate_fn,
            timeout=86400,
            # Batch size might be in kwargs!
            **kwargs,
        )
        logging.info(f"Preparing batchiterator with {dataloader_settings=}")
        prepare_dataloader = DataLoader(
            UntransformedDS(self) if get_untransformed_cases else self,
            **dataloader_settings,
        )
        return prepare_dataloader

    def prefill_cache(self):
        missing = self.get_free_cache_case_slots()
        if missing > 0:
            logging.info(f"Preparing cache of {self.cache_folder}")
            prepare_dataloader = self.get_mp_batchiterator(persistent_workers=False)
            pbar = tqdm(prepare_dataloader)
            for _ in pbar:
                missing = self.get_free_cache_case_slots()
                pbar.set_description(f"{missing} cases still missing for caching")
                if missing <= 0:
                    break

            if prepare_dataloader._iterator is not None:
                # Force kill all workers of the dataloader, just to be sure
                prepare_dataloader._iterator._shutdown_workers()  # type: ignore
                prepare_dataloader._iterator = None

    def get_free_cache_case_slots(self) -> int:
        if self.cache_folder is not None:
            assert self.cache_folder.exists()
            cached_num = len(list(self.cache_folder.glob("*.pkl")))
            # For now the number of free cache slots is exactly
            return len(self) - cached_num
        else:
            return 0

    def reset_indices(self) -> None:
        if self.reduced_size < 1.0:
            logging.warning(f"Resetting indices of {self} to full size")
        self.reduced_size = 1.0
        if self.get_dataset_style() is DatasetStyle.MapStyle:
            self._indices = np.array(list(range(len(self.src_ds))))  # type: ignore
        else:
            from mmm.torch_ext import CachingSubCaseDS

            if hasattr(self.src_ds, "supercase_ds"):
                self.src_ds.supercase_ds = self.fullcachingsubcasedssrc
            else:
                raise NotImplementedError

    def set_indices_by_fraction(self, fraction: float, seed: int = 13) -> None:
        """
        Starts by adding one example per class, then draws random samples until the fraction criterion is met.
        """
        assert fraction > 0.0 and fraction <= 1.0
        from mmm.data_loading.utils import train_val_split, TransformedSubset
        from mmm.torch_ext import CachingSubCaseDS

        self.reduced_size = fraction

        if self.get_dataset_style() is DatasetStyle.MapStyle:
            self.reset_indices()
            new_indices, _ = train_val_split(list(range(len(self.src_ds))), perc=fraction, seed=seed)  # type: ignore
            self._indices = np.array(new_indices)
        else:
            # In general, this is not possible. However, for some special case we can do it
            if hasattr(self.src_ds, "supercase_ds"):
                self.src_ds: CachingSubCaseDS
                self.fullcachingsubcasedssrc = self.src_ds.supercase_ds
                new_indices, _ = train_val_split(
                    list(range(len(self.src_ds.supercase_ds))), perc=fraction, seed=seed  # type: ignore
                )
                self.src_ds.supercase_ds = TransformedSubset(self.src_ds.supercase_ds, indices=new_indices)
            else:
                raise NotImplementedError

    def verify_all_cases(self):
        from tqdm.auto import tqdm

        for case_index in tqdm(range(len(self))):
            self.verify_case_by_index(case_index)

    def verify_case_by_index(self, index: int) -> SrcCaseType:
        d = self.get_untransformed_case(index) if self.cache_folder is None else self.get_case_from_cache(index)
        self.verify_case(d)
        return d

    def verify_case(self, d: SrcCaseType) -> None:
        missing_keys = [k for k in self.mandatory_case_keys if k not in d]
        if missing_keys:
            raise InvalidCaseError(self, -1, [f"{k} missing in case {d}" for k in missing_keys])

    def __len__(self) -> int:
        return len(self._indices)

    def __iter__(self, apply_batchtransform: bool = True) -> Iterator[SrcCaseType]:
        for d in iter(self.src_ds):
            if self.src_transform:
                d = self.src_transform(d)

            d = self.data_stripper(d)  # type: ignore

            if self.batch_transform is not None and apply_batchtransform:
                d = self.batch_transform(d)

            yield d

    def get_untransformed_case(self, index: int) -> Any:
        case = self.src_ds[self._indices[index]]
        if self.src_transform is not None:
            case = self.src_transform(case)
        return self.data_stripper(case)  # type: ignore

    def get_case_from_cache(self, index: int) -> Any:
        assert self.cache_folder is not None
        filepath = self.cache_folder / f"{self._indices[index]}.pkl"
        if filepath.exists():
            with open(filepath, "rb") as f:
                return torch.load(f)
        else:
            src_case = self.get_untransformed_case(index)
            with open(filepath, "wb") as f:
                torch.save(src_case, f)
            return src_case

    def __getitem__(self, index: int) -> SrcCaseType:
        src_case = self.get_untransformed_case(index) if self.cache_folder is None else self.get_case_from_cache(index)

        # All non deterministic transforms should be in the batch_transform
        if self.batch_transform is not None:
            src_case = self.batch_transform(src_case)

        return src_case

    def get_input_output_tuple(self, batch: SrcCaseType) -> Tuple[Any, ...]:
        """
        Relevant for ONNX and transforming cohorts to sklearn formats
        """
        raise NotImplementedError

    @abstractmethod
    def _visualize_batch_case(self, batch: SrcCaseType, i: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def _compute_batchsize_from_batch(self, batch: SrcCaseType) -> int:
        raise NotImplementedError

    def visualize_batch(self, batch: SrcCaseType) -> None:
        import streamlit as st

        batch_size = self._compute_batchsize_from_batch(batch)
        n_examples = min(8, batch_size)
        st.title(f"Drawing {n_examples} random examples from a batch with size {batch_size}")

        for i in random.sample(list(range(batch_size)), n_examples):
            self._visualize_batch_case(batch, i)

    def __repr__(self) -> str:
        res = f"MTLDataset of style {self.get_dataset_style()}"
        res += f" with mandatory keys for each case: {self.mandatory_case_keys}"
        if self.get_dataset_style is DatasetStyle.MapStyle:
            res += f" with {self.__len__()} cases"
        return res

    def st_find_invalid_cases(self) -> None:
        from mmm.logging.st_ext import stw, st

        with st.form("validator settings"):
            shuffle = st.checkbox("shuffle", value=True)

            submit = st.form_submit_button("Find invalid cases")

        if submit:
            mp_iterator = self.get_mp_batchiterator(shuffle=shuffle)
            mp_iter, i = iter(mp_iterator), 0
            progbar = st.progress(value=0.0)
            error_cases = []
            while True:
                try:
                    batch = next(mp_iter)
                    progbar.progress(
                        i / len(mp_iterator),
                        text=f"Checking step {i} of {len(mp_iterator)}",
                    )
                    i += 1
                except StopIteration:
                    st.balloons()
                    if error_cases:
                        st.error(f"Found {[x[0] for x in error_cases]} invalid cases")
                    return error_cases
                except Exception as e:
                    error_cases.append((i, e))
                    st.error(f"Could not load batch {i} due to {e}")
                    logging.error(f"Could not load batch {i} due to {e}")
                    stw(batch)

    def _st_repr_(self, st_prefix: str = "") -> None:
        from mmm.logging.st_ext import stw, st

        stw(f"### {self.__class__.__name__} of style {self.get_dataset_style()}")

        def dataset_explorer():
            if self.get_dataset_style() is DatasetStyle.MapStyle:
                case_index = st.number_input(
                    f"Select case between 0 and {len(self)}",
                    step=1,
                    value="min",
                    min_value=0,
                    max_value=len(self) - 1,
                    key=f"case_index{st_prefix}",
                )
                try:
                    case = self.verify_case_by_index(case_index)
                except InvalidCaseError as e:
                    st.error(e)
                    case = self.get_untransformed_case(case_index)
                self.st_case_viewer(case, case_index)
            else:
                with st.form("iterator demo"):
                    max_items = int(st.number_input("Max iterations", step=1, value=10))
                    display_every = int(st.number_input("Display every N case", step=1, value=2))
                    submitted = st.form_submit_button("Reload iterator")

                    if submitted:
                        for i, case in enumerate(self.__iter__(apply_batchtransform=False)):
                            self.verify_case(case)
                            # case = next(iter(self))
                            if i % display_every == 0:
                                self.st_case_viewer(case, i)
                            if i > max_items:
                                break

        def batch_explorer():
            with st.form("Dataloader settings"):
                seed = st.number_input("random_seed", value=0)
                batchsize = st.number_input("batchsize", value=1, min_value=1, max_value=100)
                shuffle: bool = st.toggle("shuffle", value=True)
                num_workers: int = st.number_input("num_workers", value=0, min_value=0, max_value=10)
                pin_memory: bool = st.toggle("pin_memory", value=False)
                submitted = st.form_submit_button("Load batches")
            if submitted:
                random.seed(seed)
                torch.manual_seed(seed)
                dataloader = self.get_dataloader(
                    batch_size=batchsize,
                    shuffle=shuffle,
                    num_workers=num_workers,
                    pin_memory=pin_memory,
                )
                batch = next(iter(dataloader))
                self.visualize_batch(batch)

        pages = {
            "Dataset explorer": dataset_explorer,
            "Batch explorer": batch_explorer,
            "Validator": self.st_find_invalid_cases,
        }

        demo_name = st.sidebar.selectbox("Choose", list(pages.keys()))
        pages[demo_name]()

    def st_case_viewer(self, case: SrcCaseType, i: int = -1) -> None:
        """
        Can be called by GUIs to visualize a specific case
        """
        from mmm.logging.st_ext import stw

        stw("Overwrite this method to visualize a case")
        stw(case)



from typing import Any
from pydantic_settings import BaseSettings

try:
    from label_studio_sdk import Client, Project
    from label_studio_converter.brush import (
        decode_rle,
        encode_rle,
        decode_from_annotation,
    )
except ImportError:
    Client, Project = Any, Any


class LabelstudioCredentials(BaseSettings):
    url: str
    token: str

    class Config:
        env_prefix = "mtltorchlabelstudio_"

    def build_client(self) -> Client:
        return Client(self.url, self.token)



import logging
from pathlib import Path
import random
import json

from pydantic import Field
import torch
import torch.nn as nn
from typing import Any, Dict, Tuple, List, cast, Optional

import numpy as np
import wandb
from mmm.logging.type_ext import StepFeedbackDict, StepMetricDict
from .MTLTask import MTLTask
from mmm.data_loading.DetectionDataset import DetectionDataset, eval_map_batch
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules

try:
    from mmdet.models.utils import multi_apply
    from mmdet.models.dense_heads import FCOSHead
    from mmengine.structures import InstanceData
    from mmengine.config import ConfigDict
except ImportError:
    logging.warning("Detection extra dependencies not installed")
    FCOSHead = nn.Module
    InstanceData = Any


class MTLFCOSHead(FCOSHead):
    #     def _init_predictor(self) -> None:
    #         """Initialize predictor layers of the head."""
    #         self.conv_cls = nn.Conv2d(
    #             self.feat_channels, self.cls_out_channels, 3, padding=1)
    #         self.conv_reg = nn.Conv2d(self.feat_channels, 4, 3, padding=1)

    def forward_single(self, cls_feat, reg_feat, scale, stride: int):
        for cls_layer in self.cls_convs:
            cls_feat = cls_layer(cls_feat)
        cls_score = self.conv_cls(cls_feat)

        for reg_layer in self.reg_convs:
            reg_feat = reg_layer(reg_feat)
        bbox_pred = self.conv_reg(reg_feat)
        # cls_score, bbox_pred, cls_feat, reg_feat = AnchorFreeHead.forward_single(self, x)

        if self.centerness_on_reg:
            centerness = self.conv_centerness(reg_feat)
        else:
            centerness = self.conv_centerness(cls_feat)
        # scale the bbox_pred of different level
        # float to avoid overflow when enabling FP16
        bbox_pred = scale(bbox_pred).float()
        if self.norm_on_bbox:
            # bbox_pred needed for gradient computation has been modified
            # by F.relu(bbox_pred) when run with PyTorch 1.10. So replace
            # F.relu(bbox_pred) with bbox_pred.clamp(min=0)
            bbox_pred = bbox_pred.clamp(min=0)

            # THE ONLY CHANGE COMPARED TO THE ORIGINAL FCOS HEAD's implementation
            # if not self.training:
            #     bbox_pred *= stride
        else:
            bbox_pred = bbox_pred.exp()
        return cls_score, bbox_pred, centerness

    def forward(self, cls_feats, reg_feats):
        """Forward features from the upstream network.

        Args:
            feats (tuple[Tensor]): Features from the upstream network, each is
                a 4D-tensor.

        Returns:
            tuple: A tuple of each level outputs.

            - cls_scores (list[Tensor]): Box scores for each scale level, \
            each is a 4D-tensor, the channel number is \
            num_points * num_classes.
            - bbox_preds (list[Tensor]): Box energies / deltas for each \
            scale level, each is a 4D-tensor, the channel number is \
            num_points * 4.
            - centernesses (list[Tensor]): centerness for each scale level, \
            each is a 4D-tensor, the channel number is num_points * 1.
        """
        return multi_apply(self.forward_single, cls_feats, reg_feats, self.scales, self.strides)

    def loss_by_feat(
        self,
        cls_scores: List[torch.Tensor],
        bbox_preds: List[torch.Tensor],
        centernesses: List[torch.Tensor],
        batch_gt_instances: list[InstanceData],
    ) -> Dict[str, torch.Tensor]:
        """
        Uses DDP blocking, but we don't share the tasks. Remove the blocking.

        So, remove the calls to "reduce_mean".
        Problematically, this method might diverge from the original implementation in future updates without noticing.
        If you know a better way to remove the reduce_mean calls, please update.
        """
        assert len(cls_scores) == len(bbox_preds) == len(centernesses)
        featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
        all_level_points = self.prior_generator.grid_priors(
            featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
        )
        labels, bbox_targets = self.get_targets(all_level_points, batch_gt_instances)

        num_imgs = cls_scores[0].size(0)
        # flatten cls_scores, bbox_preds and centerness
        flatten_cls_scores = [
            cls_score.permute(0, 2, 3, 1).reshape(-1, self.cls_out_channels) for cls_score in cls_scores
        ]
        flatten_bbox_preds = [bbox_pred.permute(0, 2, 3, 1).reshape(-1, 4) for bbox_pred in bbox_preds]
        flatten_centerness = [centerness.permute(0, 2, 3, 1).reshape(-1) for centerness in centernesses]
        flatten_cls_scores = torch.cat(flatten_cls_scores)
        flatten_bbox_preds = torch.cat(flatten_bbox_preds)
        flatten_centerness = torch.cat(flatten_centerness)
        flatten_labels = torch.cat(labels)
        flatten_bbox_targets = torch.cat(bbox_targets)
        # repeat points to align with bbox_preds
        flatten_points = torch.cat([points.repeat(num_imgs, 1) for points in all_level_points])

        # FG cat_id: [0, num_classes -1], BG cat_id: num_classes
        bg_class_ind = self.num_classes
        pos_inds = ((flatten_labels >= 0) & (flatten_labels < bg_class_ind)).nonzero().reshape(-1)
        num_pos = torch.tensor(len(pos_inds), dtype=torch.float, device=bbox_preds[0].device)
        num_pos = max(num_pos, 1.0)
        loss_cls = self.loss_cls(flatten_cls_scores, flatten_labels, avg_factor=num_pos)

        pos_bbox_preds = flatten_bbox_preds[pos_inds]
        pos_centerness = flatten_centerness[pos_inds]
        pos_bbox_targets = flatten_bbox_targets[pos_inds]
        pos_centerness_targets = self.centerness_target(pos_bbox_targets)
        # centerness weighted iou loss
        centerness_denorm = max(pos_centerness_targets.sum().detach(), 1e-6)

        if len(pos_inds) > 0:
            pos_points = flatten_points[pos_inds]
            pos_decoded_bbox_preds = self.bbox_coder.decode(pos_points, pos_bbox_preds)
            pos_decoded_target_preds = self.bbox_coder.decode(pos_points, pos_bbox_targets)
            loss_bbox = self.loss_bbox(
                pos_decoded_bbox_preds,
                pos_decoded_target_preds,
                weight=pos_centerness_targets,
                avg_factor=centerness_denorm,
            )
            loss_centerness = self.loss_centerness(pos_centerness, pos_centerness_targets, avg_factor=num_pos)
        else:
            loss_bbox = pos_bbox_preds.sum()
            loss_centerness = pos_centerness.sum()

        return dict(loss_cls=loss_cls, loss_bbox=loss_bbox, loss_centerness=loss_centerness)


class MMDetectionTask(MTLTask):
    """
    Requires to add mmcv and detection extra dependencies.
    Depending on your environment, a statement like could be added to your pyproject.toml:
    `mmcv = { url = "https://download.openmmlab.com/mmcv/dist/cu121/torch2.1.0/mmcv-2.1.0-cp310-cp310-manylinux1_x86_64.whl", optional = true }`

    The confidence score of a box is influenced by the classification output and the centerness output.
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        decoder_key: str = "fcosfpn"
        squeezer_key: str = "squeezer"
        min_threshold_for_metrics: float = Field(
            default=0.05,
            description="Threshold after which confidence score boxes are considered in metric computation",
        )
        max_boxes: int = 100
        norm_on_bbox: bool = True
        centerness_on_reg: bool = True
        center_sampling: bool = False
        conv_bias: bool = True
        stacked_convs: int = 0
        dcn_on_last_conv: bool = False
        # head_kernel_size: int = Field(
        #     default=3,
        #     description="Boxes and classes are predicted with this kernel size."
        # )

    def __init__(
        self,
        args: Config,
        for_strides: List[int],
        in_channels: int,
        cohort: TrainValCohort[DetectionDataset],
    ) -> None:
        super().__init__(args, cohort)
        self.args: MMDetectionTask.Config
        self.class_names = cohort.datasets[0].vis_classes

        self.good_thres: Optional[float] = None

        self.strides = for_strides
        logging.debug(f"Setting up mmDetection task, fpn strides: {for_strides}")

        # I think feature channels are configurable, in_channels need to be the same as the output of the neck
        fcos_head: MTLFCOSHead = MTLFCOSHead(
            num_classes=len(self.class_names),
            in_channels=in_channels,
            stacked_convs=args.stacked_convs,
            feat_channels=in_channels,
            strides=for_strides,
            loss_cls=dict(
                type="FocalLoss",
                use_sigmoid=True,
                gamma=2.0,
                alpha=0.25,
                loss_weight=1 / np.log(len(self.class_names)),
            ),
            loss_bbox=dict(type="IoULoss", loss_weight=1.0),
            loss_centerness=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=0.2),
            # Training tricks not mentioned in the original paper:
            norm_on_bbox=args.norm_on_bbox,
            centerness_on_reg=args.centerness_on_reg,
            dcn_on_last_conv=args.dcn_on_last_conv,
            center_sampling=args.center_sampling,
            conv_bias=args.conv_bias,
        )
        self.test_cfg = ConfigDict(
            nms_pre=1000,
            min_bbox_size=0,
            score_thr=0.05,
            nms=dict(type="nms", iou_threshold=0.5),
            max_per_img=args.max_boxes,
        )

        self.task_modules: nn.ModuleDict = nn.ModuleDict(
            {
                # # The same convolution applied to each classification feature map.
                # # In consequence, FPN has same channel everywhere.
                "head": fcos_head
            }
        )
        # The normalization factor might be computable

    def get_head(self) -> MTLFCOSHead:
        return cast(MTLFCOSHead, self.task_modules["head"])

    def forward(self, x: Any, shared_blocks: Dict[str, SharedBlock]):
        pyr = shared_blocks[self.args.encoder_key](x)

        # for Backward compatibility
        if hasattr(self.args, "squeezer_key") and self.args.squeezer_key in list(shared_blocks.keys()):
            pyr[-1], _ = shared_blocks[self.args.squeezer_key](pyr)

        cls_features, reg_features = shared_blocks[self.args.decoder_key](pyr)
        cls_score, bbox_pred, centerness = self.task_modules["head"].forward(cls_features, reg_features)
        return cls_score, bbox_pred, centerness

    def prepare_batch(self, batch: List[Dict[str, Any]]) -> Any:
        for d in batch:
            d["image"] = d["image"].to(self.torch_device)
            d["boxes"] = d["boxes"].to(self.torch_device)
            d["labels"] = d["labels"].to(self.torch_device)
        return batch

    def features_to_boxes(
        self, cls_score, bbox_pred, centerness, img_metas
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        img_metas is a list with the mmdet image metadata for each input image:

        img_shape without channel dimension (e.g. (512, 512))
        scale_factor (e.g. 1.)
        """
        results_list = self.get_head().predict_by_feat(
            cls_score,
            ([box * s for box, s in zip(bbox_pred, self.strides)] if self.get_head().norm_on_bbox else bbox_pred),
            centerness,
            batch_img_metas=img_metas,
            rescale=False,
            cfg=self.test_cfg,
        )
        return results_list  # type: ignore

    def training_step(
        self, batch: List[Dict[str, Any]], shared_blocks: SharedModules
    ) -> Tuple[torch.Tensor, StepFeedbackDict]:
        images, metas = [], []
        targets: List[InstanceData] = []
        for d in batch:
            images.append(d["image"])
            if d["boxes"].shape[0] > 0:
                boxes_tensor = torch.clone(d["boxes"])
            else:
                boxes_tensor = torch.empty((0, 4)).float().to(d["image"].device)
            label_tens = d["labels"] if d["boxes"].shape[0] > 0 else torch.empty((0)).long().to(d["image"].device)
            targets.append(
                InstanceData(
                    bboxes=boxes_tensor,
                    labels=label_tens,
                    metainfo={"img_shape": d["image"].shape[1:], "scale_factor": 1.0},
                )
            )
            metas.append(d["meta"] if "meta" in d else {})

        cls_score, bbox_pred, centerness = shared_blocks.forward(torch.stack(images), self.forward)

        losses: Dict = self.get_head().loss_by_feat(
            cls_scores=cls_score,
            bbox_preds=bbox_pred,
            centernesses=centerness,
            batch_gt_instances=targets,
            # batch_img_metas=[],  # Unused in their implementation of FCOS [d.img_meta for d in targets]
            # batch_gt_instances_ignore=None
        )
        cls_loss, box_loss, centerness_loss = (
            losses["loss_cls"],
            losses["loss_bbox"],
            losses["loss_centerness"],
        )
        final_loss: torch.Tensor = cls_loss + box_loss + centerness_loss

        # Compute user feedback
        real_time_feedback: StepFeedbackDict = {k: lossval.item() for k, lossval in losses.items()}  # type: ignore

        with torch.no_grad():
            boxes_foreach_image = self.features_to_boxes(
                cls_score, bbox_pred, centerness, [d.metainfo for d in targets]
            )  # scores, boxes, labels
            # import pdb
            # pdb.set_trace()
            vis_n = min(self._takeout_vis_budget(), len(images))
            if vis_n > 0:
                img_index = random.randint(0, len(images) - 1)

                # boxes_foreach_image[img_index]
                wandb_img = self.visualize_prediction(
                    images[img_index].cpu(),
                    targets[img_index].bboxes.cpu(),
                    targets[img_index].labels.cpu(),
                    boxes_foreach_image[img_index].bboxes.cpu(),
                    boxes_foreach_image[img_index].scores.cpu(),
                    boxes_foreach_image[img_index].labels.cpu(),
                    metas[img_index],
                )
                real_time_feedback["preds"] = wandb_img

            filter_by_score = [
                img_preds_bboxes.scores > self.args.min_threshold_for_metrics
                for img_preds_bboxes in boxes_foreach_image
            ]

            pred_boxes = [
                img_preds_bboxes.bboxes[boxfilter, ...].clone().cpu().numpy()
                for boxfilter, img_preds_bboxes in zip(filter_by_score, boxes_foreach_image)
            ]
            pred_scores = [
                img_preds_bboxes.scores[boxfilter].clone().cpu().numpy()
                for boxfilter, img_preds_bboxes in zip(filter_by_score, boxes_foreach_image)
            ]
            pred_labels = [
                img_preds_bboxes.labels[boxfilter].clone().cpu().numpy()
                for boxfilter, img_preds_bboxes in zip(filter_by_score, boxes_foreach_image)
            ]

            step_metrics: StepMetricDict = {
                "gtboxes": [d.bboxes.clone().cpu().numpy().astype(np.int64) for d in targets],
                "gtlabels": [d.labels.clone().cpu().numpy() for d in targets],
                "predboxes": pred_boxes,
                "predscores": pred_scores,
                "predlabels": pred_labels,
            }  # type: ignore
        self.add_step_result(final_loss.item(), step_metrics)
        return final_loss, real_time_feedback

    @torch.no_grad()
    def visualize_prediction(
        self,
        img: torch.Tensor,
        img_boxes: List,
        img_labels: List,
        pred_boxes: List,
        pred_scores: List,
        pred_labels: List,
        meta: Dict,
    ) -> wandb.Image:
        gt_boxes = [
            {
                "position": {
                    "minX": int(box[0]),
                    "maxX": int(box[2]),
                    "minY": int(box[1]),
                    "maxY": int(box[3]),
                },
                "domain": "pixel",
                "class_id": int(box_label),
            }
            for box, box_label in zip(img_boxes, img_labels)
        ]

        vis_boxes = [
            {
                "position": {
                    "minX": int(box[0]),
                    "maxX": int(box[2]),
                    "minY": int(box[1]),
                    "maxY": int(box[3]),
                },
                "domain": "pixel",
                "class_id": int(box_label),
                "scores": {"score": box_score.item()},
                # "box_caption": "test_caption"
            }
            for box, box_score, box_label in zip(pred_boxes, pred_scores, pred_labels)
        ]

        metastr = json.dumps(meta, default=lambda o: str(o))
        im = wandb.Image(
            img,
            boxes={
                "predictions": {
                    "class_labels": {i: f"{v}_pred" for i, v in enumerate(self.class_names)},
                    "box_data": vis_boxes,
                },
                "ground_truth": {
                    "class_labels": {i: f"{v}_gt" for i, v in enumerate(self.class_names)},
                    "box_data": gt_boxes,
                },
            },
            caption=f"{np.min(img.numpy()):.3f}, {np.max(img.numpy()):.3f}\n{img.shape}\n{metastr}",
        )
        return im

    def determine_thresholds(self, proposal_thresholds: List[float]):
        # threshold -> mAP
        res: dict[float, tuple[float, Any]] = {}
        for t in proposal_thresholds:
            predboxes = [
                np.hstack([m[score > t], score[score > t].reshape(-1, 1)])
                for metrics in self._step_metrics
                for score, m in zip(metrics["predscores"], metrics["predboxes"])
            ]
            predlabels = [
                m[score > t]
                for metrics in self._step_metrics
                for score, m in zip(metrics["predscores"], metrics["predlabels"])
            ]
            gtboxes = [m for metrics in self._step_metrics for m in metrics["gtboxes"]]
            gtlabels = [m for metrics in self._step_metrics for m in metrics["gtlabels"]]

            mean_ap, details = eval_map_batch(predboxes, predlabels, gtboxes, gtlabels, self.class_names)
            res[t] = mean_ap, details
        return res

    def save_checkpoint(self, folder_path: Path):
        super().save_checkpoint(folder_path)
        meta_dict = {}
        if self.good_thres is not None:
            # Save the best threshold as json to folder_path
            meta_dict["threshold"] = self.good_thres
        with open(folder_path / "meta.json", "w") as f:
            json.dump(meta_dict, f)

    def load_checkpoint(self, folder_path: Path):
        super().load_checkpoint(folder_path)
        with open(folder_path / "meta.json", "r") as f:
            meta = json.load(f)
        if "threshold" in meta:
            self.good_thres = meta["threshold"]

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        # Build one huge "batch" with all results of the epoch and compute metrics:

        metrics, logstring = super().log_epoch_metrics()
        # assert self.args.score_threshold_for_metrics is not None

        # if self.training:
        #     if self.good_thres is None:
        #         thresholds = self.determine_thresholds(list(np.arange(self.args.min_threshold_for_metrics, 0.9, 0.1)))
        #     else:
        #         proposals = list(filter(lambda x: 0.95 >= x >= self.args.min_threshold_for_metrics, [
        #             self.good_thres - 0.05,
        #             self.good_thres,
        #             self.good_thres + 0.05,
        #         ]))
        #         thresholds = self.determine_thresholds(proposals)

        #     best_threshold = max(thresholds, key=lambda k: thresholds[k][0])
        #     self.good_thres = best_threshold
        # else:
        #     if self.good_thres is None:
        #         thresholds = self.determine_thresholds([self.args.min_threshold_for_metrics])
        #     else:
        #         thresholds = self.determine_thresholds([self.good_thres])

        thresholds = self.determine_thresholds([self.args.min_threshold_for_metrics])
        best_threshold = max(thresholds, key=lambda k: thresholds[k][0])
        metrics[f"mAP"] = thresholds[best_threshold][0]
        metrics[f"best_threshold"] = best_threshold
        logstring = f"{logstring} - map{best_threshold:.2f} {thresholds[best_threshold][0]:.3f}"
        for class_ap, class_name in zip(thresholds[best_threshold][1], self.class_names):
            metrics[f"{class_name}AP"] = class_ap["ap"]

        return metrics, logstring



from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.tasks.ClassificationTask import ClassificationTask
from mmm.mtl_modules.tasks.SemSegTask import SemSegTask, MultiLabelSemSegTask
from mmm.mtl_modules.tasks.RegressionTask import RegressionTask
from mmm.mtl_modules.tasks.mmDetectionTask import MMDetectionTask
from mmm.mtl_modules.tasks.MultilabelClassificationTask import MultilabelClassificationTask
from mmm.mtl_modules.tasks.ImageGenerationTask import ImageGenerationTask



import torch
import random
import colorsys
import cv2
import numpy as np
import base64


def get_colors(num_colors):
    colors = []
    for i in np.arange(0.0, 360.0, 360.0 / num_colors):
        hue = i / 360.0
        lightness = (50 + np.random.rand() * 10) / 100.0
        saturation = (90 + np.random.rand() * 10) / 100.0
        colors.append(colorsys.hls_to_rgb(hue, lightness, saturation))
    return colors


def rgbnumpy_to_base64(img: np.ndarray) -> str:
    """
    Expects a numpy array with shape (H, W, C) and converts it to a base64 string
    """
    _, buffer = cv2.imencode(".jpg", cv2.cvtColor(img, cv2.COLOR_RGB2BGR))
    encoded_img = buffer.tobytes()
    base64_img_string = base64.b64encode(encoded_img).decode("utf-8")
    return f"data:image/jpeg;base64,{base64_img_string}"



"""
Wraps an nn.ModuleDict that was created using `trainer.save_blocks_native(...)`.
"""

import logging
from pathlib import Path
import torch
import torch.nn as nn

from mmm.data_loading.DistributedPath import DistributedPath
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.utils import get_default_cachepath

MMM_MODELS = {"encoder-1.0.3.pt.zip": "https://owncloud.fraunhofer.de/index.php/s/lIS1DlZ0BFPhFAb/download"}
DEFAULT_MODEL = "encoder-1.0.3.pt.zip"


class NativeBlocks:
    """
    Wraps an nn.ModuleDict that was created using `trainer.save_blocks_native(...)`.
    """

    def __init__(self, modules_path: DistributedPath | str, device_identifier: str = "cuda") -> None:
        if isinstance(modules_path, str):
            modules_path = DistributedPath(uri=modules_path)
        if modules_path.uri.startswith("http"):
            # Download file to cache
            cache_folder = get_default_cachepath("models")
            model_name = {v: k for k, v in MMM_MODELS.items()}.get(modules_path.uri, Path(modules_path.uri).name)
            if not model_name.endswith(".pt.zip"):
                model_name = f"{model_name}.pt.zip"
            if not (target_path := cache_folder / model_name).exists():
                logging.info(f"Model does not exist at {target_path}, downloading...")
                torch.hub.download_url_to_file(modules_path.uri, target_path)
            else:
                logging.info(f"Model already exists at {target_path}, not redownloading")
            modules_path = DistributedPath(uri=str(target_path), options=modules_path.options)

        self.device = device_identifier
        if modules_path.uri.endswith(".pt.zip"):
            import zipfile
            import tempfile

            # Extract the zip file to a temporary directory
            with zipfile.ZipFile(modules_path.upath(), "r") as zip_ref:
                with tempfile.TemporaryDirectory() as tmpdirname:
                    zip_ref.extractall(tmpdirname)
                    modules_path = DistributedPath(uri=tmpdirname, options=modules_path.options)
                    # Load the modules from the temporary directory
                    self.torch_modules: nn.ModuleDict = nn.ModuleDict()
                    for file in modules_path.upath().glob("*.pt"):
                        try:
                            self.torch_modules[file.stem] = torch.load(file).to(device_identifier)
                        except Exception as e:
                            logging.error(f"Did not load {file} due to {e}")

        else:
            self.torch_modules = torch.load(modules_path.upath()).to(device_identifier)

    def get_sharedblock_keys(self) -> list[str]:
        return [k for k, v in self.torch_modules.items() if isinstance(v, SharedBlock)]

    def get_task_keys(self) -> list[str]:
        return [k for k, v in self.torch_modules.items() if isinstance(v, MTLTask)]

    def get_device(self) -> str:
        return self.device

    def __getitem__(self, key):
        return self.torch_modules[key]

    def keys(self):
        return self.torch_modules.keys()

    @staticmethod
    def save_to_disk(file_path: DistributedPath, d: nn.ModuleDict):
        """
        Stores the modules in a folder with one file per module, where the filename is the f"{key}.pt".
        The filename could be the unique name returned by `module.get_name()`.
        """
        if file_path.uri.endswith(".pt.zip"):
            import zipfile
            import tempfile

            with zipfile.ZipFile(file_path.upath(), "w") as zip_ref:
                for k, v in d.items():
                    with tempfile.NamedTemporaryFile() as tmpfile:
                        torch.save(v, tmpfile.name)
                        zip_ref.write(tmpfile.name, f"{k}.pt")
                return file_path
        else:
            # Save the whole dict to a single file
            torch.save(d, file_path.upath())



from enum import Enum
from typing import Callable
import torch.nn as nn

from mmm.neural.TorchModule import TorchModule


class GlobalPooling(str, Enum):
    AveragePooling = "average"
    MaxPooling = "max"


pooling_converter = {
    GlobalPooling.AveragePooling: nn.AdaptiveAvgPool2d,
    GlobalPooling.MaxPooling: nn.AdaptiveMaxPool2d,
}


class GlobalPoolingConfig(TorchModule):
    pooling_type: GlobalPooling

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return pooling_converter[self.pooling_type]((1, 1))



"""
Modules wrapped from the segmentation models for PyTorch library
"""

from __future__ import annotations
import logging
from typing import Literal
import torch
import torch.nn as nn
from pydantic import Field

from segmentation_models_pytorch.decoders.unet.decoder import UnetDecoder
from segmentation_models_pytorch.decoders.pan.decoder import PANDecoder

from ..TorchModule import TorchModule
from ..activations import ActivationFunctionConfig, ActivationFn


class SMPUnetDecoder(nn.Module):
    """
    Wraps segmentation-models-pytorch's UnetDecoder

    >>> import torch
    >>> from mmm.neural.modules.smp_modules import SMPUnetDecoder
    >>> B, out_channels_for_pixel, pyr_depth, pyr_stride = 4, 16, (3, 64, 256, 512, 1024, 2048), (1, 2, 4, 8, 16, 32)
    >>> test_pyramid = [torch.rand((B, d, 224 // s, 224 // s)) for d, s in zip(pyr_depth, pyr_stride)]
    >>> dec = SMPUnetDecoder(SMPUnetDecoder.Config(pixel_embedding_dim=out_channels_for_pixel), pyr_depth, pyr_stride[-1])
    >>> dec(test_pyramid).shape
    torch.Size([4, 16, 224, 224])
    >>> assert True not in [c > dec.args.hidden_dim for c in dec.channels]  # use hidden_dim as the max number of channels
    """

    class Config(TorchModule):
        architecture: Literal["unetdecoder"] = "unetdecoder"
        pixel_embedding_dim: int = 16
        hidden_dim: int = Field(96, description="The maximum depth of the decoder. Avoids excessive memory usage.")
        head_kernel_size: int = 3
        activation: ActivationFunctionConfig = ActivationFunctionConfig(fn_type=ActivationFn.GeLU)

        def build_instance(self, *args, **kwargs) -> SMPUnetDecoder:
            return SMPUnetDecoder(self, enc_out_channels=args[0], encoder_output_stride=args[1])

    def __init__(self, args: Config, enc_out_channels: list[int], encoder_output_stride: int) -> None:
        super().__init__()
        self.args = args
        # Skip one channel because that is the raw input
        self.channels = [
            min(((2**i) * args.pixel_embedding_dim), args.hidden_dim)
            for i in reversed(range(len(enc_out_channels) - 1))
        ]
        self.final_conv = nn.Conv2d(
            min(args.pixel_embedding_dim, args.hidden_dim),
            self.args.pixel_embedding_dim,
            kernel_size=self.args.head_kernel_size,
            padding=self.args.head_kernel_size // 2,
        )
        self.norm = nn.BatchNorm2d(num_features=self.args.pixel_embedding_dim)
        self.activation = args.activation.build_instance()
        assert encoder_output_stride == 32
        logging.debug(f"Channels of {self.args.architecture} are {self.channels}")
        self.decoder = UnetDecoder(enc_out_channels, tuple(self.channels), n_blocks=len(self.channels))

    def get_upsampling_factor(self) -> int:
        return 1

    def get_output_dim_per_pixel(self) -> int:
        return self.args.pixel_embedding_dim

    def forward(self, features: list[torch.Tensor]) -> torch.Tensor:
        hidden_map: torch.Tensor = self.decoder(*features)
        return self.activation(self.norm(self.final_conv(hidden_map)))


class SMPPyramidAttentionDecoderConfig(TorchModule):
    architecture: Literal["pyramid_attention"] = "pyramid_attention"
    decoder_channels: int = 32

    def build_instance(self, *args, **kwargs) -> SMPPyramidAttentionDecoder:
        return SMPPyramidAttentionDecoder(self, enc_out_channels=args[0], encoder_output_stride=args[1])


class SMPPyramidAttentionDecoder(nn.Module):
    """
    Wraps segmentation-models-pytorch's PAN decoder
    """

    def __init__(
        self,
        args: SMPPyramidAttentionDecoderConfig,
        enc_out_channels: list[int],
        encoder_output_stride: int,
    ) -> None:
        super().__init__()
        self.args = args

        assert len(enc_out_channels) > 4, "By the time of writing, PAN uses the latest four levels"

        assert encoder_output_stride == 32
        self.decoder = PANDecoder(enc_out_channels, self.args.decoder_channels)

    def get_upsampling_factor(self) -> int:
        return 4

    def get_output_dim_per_pixel(self) -> int:
        return self.args.decoder_channels

    def forward(self, features: list[torch.Tensor]):
        return self.decoder(*features)
